package com.sillycycle.bagleyd.loop;

/*
 * @(#)LoopBoxes.java
 *
 * Copyright 2010 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JComboBox;
import javax.swing.JPanel;

import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComboBox;

/**
 * The <code>LoopBoxes</code> class holds the Life configuration
 * for the <code>LoopDialog</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class LoopBoxes extends JPanel {
	private static final long serialVersionUID = 42L;

	static int neighborIndexValue = LoopInterface.DEFAULT_NEIGHBOR_INDEX;
	/* default setup so that ruleIndexValue will not be null */
	static int ruleIndexValue = LoopRules.defaultRuleIndex[neighborIndexValue];
	static String ruleName = LoopRules.getRuleName(neighborIndexValue, ruleIndexValue);
	transient LoopRules loopRules = new LoopRules(LoopRules.indexToNeighbor[neighborIndexValue], true, false);

	ComboBox neighborComboBox = new ComboBox("Neighborhood Modes",
		LoopRules.neighborIndexStrings);
	ComboBox ruleComboBox = new ComboBox("Rule",
		LoopRules.getStrings(neighborIndexValue));

	public LoopBoxes() {
		JPanel panel = new JPanel();

		panel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		neighborComboBox.setSelectedItem(neighborIndexValue);
		panel.add(neighborComboBox);
		add(BorderLayout.PAGE_START, panel);
		ruleComboBox.setSelectedItem(ruleIndexValue);
		panel.add(ruleComboBox);
		//pack();
		neighborComboBox.addActionListener(
			new NeighborComboBoxActionListener());
		ruleComboBox.addActionListener(
			new RuleComboBoxActionListener());
	}

	class NeighborComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int oldValue = neighborIndexValue;

			neighborIndexValue = neighborComboBox.getSelectedItem();
			if (neighborIndexValue != -1 && oldValue != neighborIndexValue) {
				//ruleName = LoopRules.getRuleName(neighborIndexValue);
				Integer ruleIndex = LoopRules.defaultRuleIndex[neighborIndexValue];
				//String ruleName;

				if (ruleIndex == null) {
					ruleIndexValue = 0;
					//ruleName = LoopRules.defaultRuleStrings[neighborIndexValue];
				} else {
					ruleIndexValue = ruleIndex;
					//((JComboBox)event.getSource()).setSelectedIndex(ruleIndex);
					//ruleName = LoopRules.getRuleName(neighborIndexValue, ruleIndexValue);
				}
				setRuleIndex(ruleIndexValue);
				ruleComboBox.updateList(LoopRules.getStrings(neighborIndexValue),
					ruleIndexValue);
			}
		}
	}

	class RuleComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			String[] ruleString = null;

			String menuSelection = ((String) ((JComboBox<Object>)event.
				getSource()).getSelectedItem());

			if (menuSelection != null) {
				ruleString = menuSelection.split(",");
				// non hash way
				/*for (int ruleIndexValue = 0;
						ruleIndexValue < LoopRules.loopRules[neighborIndexValue].length;
						ruleIndexValue++) {
					if (ruleString[0].equals(LoopRules.loopRules[neighborIndexValue][ruleIndexValue].getRuleName())) {*/
				Integer rule = loopRules.getRuleIndex(neighborIndexValue, ruleString[0]);
				if (rule != null) {
					ruleIndexValue = rule;
					ruleName = LoopRules.getRuleName(neighborIndexValue, ruleIndexValue);
//System.out.println("actionPerformed: " + ruleName);
					LoopDialog.ruleStateChanged(neighborIndexValue);
				}
			}
			if (ruleString == null/* || "Other".equals(ruleString[0])*/) {
				Integer ruleIndex = LoopRules.defaultRuleIndex[neighborIndexValue];

				if (ruleIndex == null) {
					ruleName = LoopRules.defaultRuleStrings[neighborIndexValue];
				} else {
					ruleIndexValue = ruleIndex;
					//((JComboBox)event.getSource()).setSelectedIndex(ruleIndex);
					ruleName = LoopRules.getRuleName(neighborIndexValue, ruleIndexValue);
				}
			}
		}
	}

	public void addActionListener(ActionListener l) {
		neighborComboBox.addActionListener(l);
		neighborIndexValue = neighborComboBox.getSelectedItem();

		//ruleName = LoopRules.getRuleName(neighborIndexValue);
		Integer ruleIndex = LoopRules.defaultRuleIndex[neighborIndexValue];

		if (ruleIndex == null) {
			ruleName = LoopRules.defaultRuleStrings[neighborIndexValue];
		} else {
			ruleIndexValue = ruleIndex;
			ruleName = LoopRules.getRuleName(neighborIndexValue, ruleIndexValue);
		}
	}

	public void setNeighborIndex(int value) {
		LoopDialog.neighborIndexValue = value;
		neighborComboBox.setSelectedItem(value);
	}

	public void setRuleIndex(int value) {
		ruleComboBox.setSelectedItem(value);
	}

	public int getNeighborIndex() {
		return neighborComboBox.getSelectedItem();
	}

	public int getRuleIndex() {
		return ruleComboBox.getSelectedItem();
	}
}
