package com.sillycycle.bagleyd.loop;

/*
 * @(#)LoopCanvas.java
 *
 * Copyright 2010 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.Random;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.util.Cell2D;
import com.sillycycle.bagleyd.util.CellInfo;
import com.sillycycle.bagleyd.util.ModArith;

/**
 * The <code>LoopCanvas</code> class does all the graphics for
 * <code>LoopFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class LoopCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");

	static final byte MAX_STATES = 9;
	static final boolean MONO = false;
	static final String DATAFILE = "data.life";

	Random generator;
	int startSpeed = LoopInterface.DEFAULT_SPEED;
	int speed = LoopInterface.DEFAULT_SPEED;
	int columns = LoopInterface.DEFAULT_COLUMNS;
	int rows = LoopInterface.DEFAULT_ROWS;
	int ncols = columns;
	int nrows = rows;
	int minCol = columns / 2 - 1;
	int maxCol = columns / 2 + 1;
	int minRow = rows / 2 - 1;
	int maxRow = rows / 2 + 1;
	boolean randomOrient = false;
	boolean grid = true;
	boolean wrap = true;
	boolean visible = true;
	boolean focus = true;
	boolean mouseDown = false;
	int xs = LoopInterface.DEFAULT_CELLSIZE;
	int ys = LoopInterface.DEFAULT_CELLSIZE;
	int neighbors =
		LoopRules.indexToNeighbor[LoopInterface.DEFAULT_NEIGHBOR_INDEX];
	Color initialCellColor = Color.red;
	Color gridColor = Color.gray;
	Color background = Color.black;
	Cell2D cellPl;
	transient Capturer capturer;
	JFrame frame = null;
	boolean debug = false;
	boolean clockwise = false;
	String name = "newbie";
	LoopMatrix field = null;
	int gen = 0;
	int nCells = 0, cCells = 0;
	String stringSave = null;
	transient LoopRules param = null;

	private synchronized void pulse(boolean draw) {
		int[] na;

		cCells = 0;
		/* 1. */
		/* for each live cell in linked list
		 *   create a dead list if toggle 0 and field array 0
		 * set toggle in field */
		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int col = getColumn(position);
			int row = getRow(position);
			for (int neighbor = 0; neighbor < Math.abs(neighbors); neighbor++)
				field.setDeadCell(positionOfNeighbor(neighbor, col, row));
		}
		/* 2. */
		/* for each dead cell: reset toggles back to 0;
		 *   if comes alive keep, else remove */
		for (Iterator<CellInfo> iterator = field.deadList.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int deadPosition = cellInfo.getPosition();
			int col = getColumn(deadPosition);
			int row = getRow(deadPosition);
			byte state = 0;

			na = getNeighborArray(col, row);
			state = param.getTransition(state, na, !clockwise);
			field.resetCellBit(deadPosition);
			if (state == 0) {
				iterator.remove();
			} else {
				cellInfo.setTemp(state);
			}
		}
		/* 3. */
		/* for each live cell set toggle if dies */
		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int col = getColumn(position);
			int row = getRow(position);
			byte nextState;
			byte state;

			na = getNeighborArray(col, row);
			state = cellInfo.getCell();
			nextState = param.getTransition(state, na, !clockwise);
			if (nextState != state)
				cellInfo.setTemp((byte) (nextState + 1)); // come back later and set cell to 0
		}
		/* 4. */
		/* for each live cell update according to toggle
		 *   if dies erase cell remove.
		 *   if lives update age, draw if necessary. */
		for (Iterator<CellInfo> iterator = field.list.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int position = cellInfo.getPosition();
			byte temp = cellInfo.getTemp();
			if (temp == 1) {
				if (draw) {
					eraseCell(cellInfo);
				}
				field.setArrayCell(position, (byte) 0);
				nCells--;
				cCells++;
				iterator.remove();
			} else if (temp != 0) {
				temp = (byte) (temp - 1);
				cellInfo.setTemp((byte) 0);
				cellInfo.setCell(temp);
				field.setArrayCell(position, temp);
				if (draw) {
					drawCell(cellInfo);
				}
				cCells++;
			}
		}
		/* 5. */
		/* for each dead cell update to live cell. */
		for (Iterator<CellInfo> iterator = field.deadList.iterator();
				iterator.hasNext();) {
			CellInfo cellInfo = iterator.next();
			int deadPosition = cellInfo.getPosition();
			int col = getColumn(deadPosition);
			int row = getRow(deadPosition);
			byte nextState = cellInfo.getTemp();
			cellInfo.setCell(nextState);
			cellInfo.setTemp((byte) 0);
			if (draw)
				drawCell(cellInfo);
			field.setCell(deadPosition, nextState);
			checkAliveBounds(col, row);
			nCells++;
			cCells++;
			iterator.remove();
		}
		if (nCells > 0) {
			gen++;
		}
	}

	void loopPulse(boolean draw, boolean fast) {
		pulse(draw);
		if (draw && (!fast || (gen & 63) == 0)) {
			((LoopFrame) frame).updateGen(gen);
			((LoopFrame) frame).updateCells(nCells);
		}
	}

	public LoopCanvas() {
		generator = new Random(System.nanoTime());
		capturer = new Capturer();
		inits();
	}

	public void initTable() {
		int neighborIndex = getNeighborIndex();
		int ruleIndex = 0;

		if ((LoopFrame) frame != null)
			ruleIndex = ((LoopFrame) frame).getRuleIndex();
		//	LoopRules.defaultRuleStings[getNeighborIndex()],
		param = new LoopRules(neighbors,
			LoopRules.getRuleName(neighborIndex, ruleIndex));
		param.initTable(neighbors, ruleIndex);
	}

	final void inits() {
		initTable();
		resizeField();
		clearField(true);
	}

	void ruleChange() {
		if ((LoopFrame) frame != null) {
			param = new LoopRules(neighbors,
				LoopRules.getRuleName(getNeighborIndex(), ((LoopFrame) frame).getRuleIndex()));
//				LoopRules.defaultRuleStrings[getNeighborIndex()]);
			String ruleName = param.getRuleName();

			if (ruleName == null)
				((LoopFrame) frame).updateRule("");
			else
				((LoopFrame) frame).updateRule(ruleName);
//			System.out.println("ruleChange: " + ruleName);
		}
	}

	public int getNeighborIndex() {
		return LoopRules.neighborToIndex(neighbors);
	}

	Integer getRuleIndex(int neighborIndex) {
		String ruleName = param.getRuleName();

		if (ruleName == null) {
			System.out.println("ruleName is null");
			return null;
		}
//		System.out.println("getRuleIndex: " + ruleName);
		return param.getRuleIndex(neighborIndex, ruleName);
	}

	static boolean polygonChange(int oldNeighborIndex, int newNeighborIndex) {
		if (oldNeighborIndex != newNeighborIndex) {
			int oldPolygon = LoopRules.neighborIndexToPolygon[oldNeighborIndex];
			int newPolygon = LoopRules.neighborIndexToPolygon[newNeighborIndex];

			return (oldPolygon != newPolygon);
		}
		return false;
	}

	int getPolygon() {
		return LoopRules.neighborToPolygon(neighbors);
	}

	void setNeighbors(int neighborIndex) {
		neighbors = LoopRules.indexToNeighbor[neighborIndex];
	}

	int getRealColumns(int cols) {
		int polygon = getPolygon();
		int realColumns = cols;

		if (polygon == 3 || polygon == 6) {
			if ((realColumns & 1) == 1)
				realColumns--;
		} else if (polygon == 5) {
			realColumns -= realColumns % 4;
		}
		return realColumns;
	}

	int getRowsFromColumns(int cols) {
		int polygon = getPolygon();
		int calculatedRows = 16 * cols / 25;

		if ((polygon == 3 || polygon == 5 || polygon == 6) &&
				((calculatedRows & 1) == 1)) {
			calculatedRows--;
		}
		return calculatedRows;
	}

	public int getColumns() {
		return this.ncols;
	}

	public int getRows() {
		return this.nrows;
	}

	public void speedUp() {
		speed += 10;
		if (speed > LoopInterface.MAX_SPEED)
			speed = LoopInterface.MAX_SPEED;
		((LoopFrame) frame).setDelay(LoopInterface.MAX_SPEED +
			LoopInterface.MIN_SPEED - 1 - speed);
	}

	public void slowDown() {
		speed -= 10;
		if (speed < LoopInterface.MIN_SPEED)
			speed = LoopInterface.MIN_SPEED;
		((LoopFrame) frame).setDelay(LoopInterface.MAX_SPEED +
			LoopInterface.MIN_SPEED - 1 - speed);
	}

	void resizeField() {
		int newWidth = getSize().width;
		int newHeight = getSize().height;
		int tempCols;
		int polygon = getPolygon();

		capturer.clearCapture();
		tempCols = LoopFrame.getColumns();
		ncols = getRealColumns(tempCols);
		nrows = getRowsFromColumns(tempCols);
		if (newWidth == 0)
			newWidth = LoopInterface.DEFAULT_CELLSIZE *
				ncols + 4;
		if (newHeight == 0)
			newHeight = LoopInterface.DEFAULT_CELLSIZE *
				nrows + 4;
		cellPl = new Cell2D(polygon, ncols, nrows, newWidth, newHeight, false);
//System.out.println("Resize " + neighbors);
//System.out.println("ncols " + ncols + ", " + nrows);
//System.out.println("cols " + columns + ", " + rows);
//System.out.println(param.getCommonName());
		if (field == null) {
			columns = ncols;
			rows = nrows;
			field = new LoopMatrix(ncols, nrows);
			gen = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			nCells = 0;
			if ((LoopFrame) frame != null) {
				((LoopFrame) frame).initPopup();
			}
		} else if (columns != ncols || rows != nrows) {
			if (polygon == 3) {
				if (((minCol + minRow) & 1) == 1) {
					if ((minRow & 1) == 1) {
						minRow--;
					} else {
						minCol--;
					}
				}
			} else if (polygon == 5) {
				int mod = (minCol + 2 * minRow) % 4;

				if ((mod / 2) == 1) {
					if ((minRow & 1) == 1) {
						minRow--;
					} else {
						minCol -= 2 ;
					}
				}
				minCol -= (mod & 1);
			} else if (polygon == 6 && (minRow & 1) == 1) {
				minRow--;
			}
			recalibrateBounds(columns, rows);

			int fieldX = maxCol - minCol + 1;
			int fieldY = maxRow - minRow + 1;
			LoopMatrix cage = new LoopMatrix(fieldX, fieldY);
			for (CellInfo cellInfo : field.list) {
				int oldP = cellInfo.getPosition();
				int i = getColumn(oldP, columns);
				int j = getRow(oldP, columns);
				byte value = field.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i - minCol, j - minRow,
						fieldX);
					cage.setArrayCell(p, value);
				}
			}
			field.clearLiveCells();
			columns = ncols;
			rows = nrows;
			field = new LoopMatrix(ncols, nrows);
			gen = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			nCells = 0;

			int offsetX = minCol + 1 - fieldX / 2;
			int offsetY = minRow + 1 - fieldY / 2;

			if (polygon == 3) {
				if (((offsetX + offsetY) & 1) == 1) {
					if ((offsetY & 1) == 1) {
						offsetY--;
					} else {
						offsetX--;
					}
				}
			} else if (polygon == 5) {
				int mod = (offsetX + 2 * offsetY) % 4;

				if ((mod / 2) == 1) {
					if ((offsetY & 1) == 1) {
						offsetY--;
					} else {
						offsetX -= 2 ;
					}
				}
				offsetX -= (mod & 1);
			} else if (polygon == 6 && (offsetY & 1) == 1) {
				offsetY--;
			}
			for (int j = 0; j < fieldY; j++)
			if ((j + offsetY >= 0) && (j + offsetY <= nrows - 1))
			for (int i = 0; i < fieldX; i++)
			if ((i + offsetX >= 0) && (i + offsetX <= ncols - 1)) {
				int oldP = getPosition(i, j, fieldX);
				byte value = cage.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i + offsetX, j + offsetY,
						ncols);
					field.setCell(p, value);
					nCells++;
				}
			}
			if ((LoopFrame) frame != null) {
				((LoopFrame) frame).updateGen(gen);
				((LoopFrame) frame).updateCells(nCells);
				((LoopFrame) frame).initPopup();
			}
		}
	}

	void xorCell(Graphics g, CellInfo info, int k) {
		int col = getColumn(info);
		int row = getRow(info);
		byte state = info.getCell();
		Color color;

		if (state == 0) {
			color = background;
		} else {
			if (MONO) {
				color = Color.white;
			} else {
				color = colorTable(state);
			}
		}
		if ((k & 1) == 0) // Simulate Xor
			color = new Color(255 - color.getRed(),
				255 - color.getGreen(),
				255 - color.getBlue());
		cellPl.drawCell(g, false, color, (grid) ? gridColor : background,
			col, row);
	}

	// Should make some attempt at minimizing size if on a border...
	void printField(StringBuilder sb) {
		String space = " ";
		String newline = NL;
		String pound = "#";
		int polygon = getPolygon();
		boolean firstPoint = true;
		int i, j, p;
		int startX = 0, startY = 0;
		int endX = 0, endY = 0;
		byte state;

		sb.append(pound).append("Dim").append(space)
			.append(2).append(newline);
		sb.append(pound).append("Neighbors").append(space)
			.append(neighbors).append(newline);
		sb.append(pound).append("Rule").append(space)
			.append(newline);
		sb.append(pound).append("Cells").append(space)
			.append(nCells).append(newline);
		sb.append(pound).append("Name").append(space)
			.append(name).append(newline);
		sb.append(pound).append("Clockwise").append(space)
			.append((clockwise) ? 1 : 0).append(newline);
		//sb.append(pound + "Sig" + space + "An A0 <> D1 Dn" + newline);
		for (j = 0; j < nrows; j++)
			for (i = 0; i < ncols; i++) {
				p = getPosition(i, j);
				if (field.getCell(p) != 0) {
					if (firstPoint) {
						startX = i;
						startY = j;
						endX = i;
						endY = j;
						firstPoint = false;
					} else {
						if (i < startX)
							startX = i;
						if (j < startY)
							startY = j;
						if (i > endX)
							endX = i;
						if (j > endY)
							endY = j;
					}
				}
			}
		sb.append(pound).append("Width").append(space)
			.append(endX - startX + 1).append(newline);
		sb.append(pound).append("Height").append(space)
			.append(endY - startY + 1).append(newline);
		sb.append(pound).append("P").append(space)
			.append(startX - ncols / 2).append(space)
			.append(startY - nrows / 2).append(newline);
		/*sb.append(pound).append("P").append(space)
			.append(0).append(space)
			.append(0).append(newline);*/
		switch (polygon) {
		case 3:
			if (((startX + startY) & 1) == 1) {
				startX--;
			}
			break;
		case 5:
			startX -= (startX + 2 * startY) % 4;
			break;
		case 6:
			int append = (endY - startY + 1) / 2;
			int low = endX + append - startX;
			int high = 0;
			int offset = 0;

			for (j = startY; j <= endY; j++) {
				for (i = startX + offset; i <= endX + append; i++) {
					p = getPosition(i, j);
					state = field.getCell(p);
					if (state != 0) {
						if (low > i - startX - offset)
							low = i - startX - offset;
						if (high < i - startX - offset )
							high = i - startX - offset;
					}
				}
				if ((j & 1) == 1) {
					offset--;
					append--;
				}
			}
			append = high;
			offset = low;
			for (j = startY; j <= endY; j++) {
				for (int k = j; k < endY; k++) {
					sb.append(space);
				}
				for (i = startX + offset; i <= startX + append; i++) {
					if (i > startX + offset)
						sb.append(space);
					p = getPosition(i, j);
					sb.append(field.getCell(p));
				}
				if ((j & 1) == 1) {
					offset--;
					append--;
				}
				sb.append(newline);
			}
			return;
		default:
			break;
		}
		for (j = startY; j <= endY; j++) {
			for (i = startX; i <= endX; i++) {
				p = getPosition(i, j);
				sb.append(field.getCell(p));
			}
			sb.append(newline);
		}
	}

	void clearField(boolean draw) {
		field.clearLiveCells();
		if (draw)
			eraseFrame();
		gen = 0;
		minCol = ncols / 2 - 1;
		maxCol = ncols / 2 + 1;
		minRow = nrows / 2 - 1;
		maxRow = nrows / 2 + 1;
		nCells = 0;
		if (draw && (LoopFrame) frame != null) {
			Graphics g = this.getGraphics();

			((LoopFrame) frame).updateGen(gen);
			((LoopFrame) frame).updateCells(nCells);
			drawFrame(g, true);
			if (grid)
				drawGrid();
		}
		prepopulate();
	}

	void prepopulate() {
		/*int polygon = getPolygon();*/
		int ni = getNeighborIndex();
		Integer ri = getRuleIndex(ni);

		initTable();
		//System.out.println("prepopulate: ni " + ni + ", ri " + ri);
		getFormFromMenu(LoopForms.getName(ni, ri, 0));
		/*if (polygon == 4) {
			getFormFromMenu("selfReproducingLoop");
		} else if (polygon == 6) {
			getFormFromMenu("hexSelfReproducingLoop");
		}*/
	}

	void showMessage(String msg) {
		((LoopFrame) frame).showMessage(msg);
	}

	void setCell(int col, int row, byte state) {
		if (col >= 0 && col < ncols &&
				row >= 0 && row < nrows) {
			setCell(col, row, state, visible);
		}
	}

	/* rules are not reflective symmetric so must know */
	private boolean mirror(int orient) {
		int polygon = getPolygon();

		if (polygon == 4) {
			if (orient >= 4) {
				if ((orient % 4 >= 2) &&
						((orient & 1) == 1))
					return true;
				if ((orient % 4 < 2) &&
						((orient & 1) == 0))
					return true;
			} else {
				if ((orient % 4 >= 2) &&
						((orient & 1) == 0))
					return true;
				if ((orient % 4 < 2) &&
						((orient & 1) == 1))
					return true;
			}
		} else if (polygon == 6 && orient >= 6) {
			return true;
		}
		return false;
	}

	private void setOrientCell(int x, int y,
			int startX, int startY, byte state,
			int orient, boolean draw) {
		int i = x, j = y, temp;
		int polygon = getPolygon();

		switch (polygon) {
			case 3:
				if (orient >= 6) {
					i = -i;
				}
				switch (orient % 6) {
				default:
				case 0:
					break;
				case 1:
					temp = j;
					j = ModArith.ceil2(temp + i);
					i = ModArith.ceil2(i - temp + 1) - temp;
					break;
				case 2:
					temp = j;
					j = ModArith.ceil2(temp - i);
					i = temp - ModArith.ceil2(1 - i - temp);
					break;
				case 3:
					i = -i;
					j = -j + 1;
					break;
				case 4:
					temp = j;
					j = 1 - ModArith.ceil2(temp + i);
					i = temp - ModArith.ceil2(i - temp + 1);
					break;
				case 5:
					temp = j;
					j = 1 - ModArith.ceil2(temp - i);
					i = ModArith.ceil2(1 - i - temp) - temp;
					break;
				}
				break;
			case 4:
				if (orient >= 4) {
					temp = i;
					i = j;
					j = temp;
				}
				if (orient % 4 >= 2) {
					i = -i;
				}
				if ((orient & 1) == 1) {
					j = -j;
				}
				break;
			case 5:
				int mod = ModArith.mod(i + 2 * j, 4);
				int ti = j * 2;
				int tj = ModArith.div(i, 2);
				if (mod == 1 || mod == 2)
					ti++;
				switch (orient % 4) {
				case 1:
					i = -ti - 1;
					j = tj;
					break;
				case 2:
					i = -i - 1;
					j = -j - 1;
					break;
				case 3:
					i = ti;
					j = -tj - 1;
					break;
				default:
					break;
				}
				if (orient >= 4) {
					i = -i + 1;
					mod = ModArith.mod(i + 2 * j, 4);
					if (mod == 0)
						i++;
					if (mod == 1)
						i--;
				}
				break;
			case 6:
				if (orient >= 6) {
					temp = i;
					i = j;
					j = temp;
				}
				switch (orient % 6) {
				default:
				case 0:
					break;
				case 1:
					temp = j;
					j = temp - i;
					i = temp;
					break;
				case 2:
					temp = -i;
					i = temp + j;
					j = temp;
					break;
				case 3:
					i = -i;
					j = -j;
					break;
				case 4:
					temp = -j;
					j = temp + i;
					i = temp;
					break;
				case 5:
					temp = i;
					i = temp - j;
					j = temp;
					break;
				}
				break;
			default:
				break;
		}
		if (polygon == 5) {
			temp = startX - (startX + startY * 2) % 4;
			i += temp;
		} else
			i += startX;
		if (polygon == 6) {
			if ((startX & 1) == 1 && j >= 0 || j > 0) {
				i += (((startX & 1) == 1) ? -j / 2 :
					-(j - 1) / 2);
			} else {
				i += 1 + (((startX & 1) == 1) ? -(j + 1) / 2 :
					-j / 2);
			}
		}
		j += startY;
		if (polygon == 3 &&
				(((startX + startY + 1) & 1) == 1)) {
			i++;
		}
		if (i >= 0 && i < ncols && j >= 0 && j < nrows)
			setCell(i, j, state, draw);
	}

	public void displayForm(byte[][] data, boolean draw) {
		int i, j;
		byte state;
		int orient = 0;
		int polygon = getPolygon();
		int startX = ncols / 2 - data[0].length / 2;
		int startY = nrows / 2 - data.length / 2;

		if (polygon == 5) {
			startX -= (startX + startY * 2) % 4;
		} else if (polygon == 6 && ((startX + startY) & 1) == 0) {
			startY -= 1;
		}
		if (randomOrient) {
			if (polygon == 3 || polygon == 6) {
				orient = generator.nextInt(12);
			} else if (polygon == 4 || polygon == 5) {
				orient = generator.nextInt(8);
			}
		}  else { /* not sure why needed */
			if (polygon == 6) {
				orient = 6;
			} else if (polygon == 4) {
				orient = 4;
			}
		}
		clockwise = !mirror(orient);
		for (j = 0; j < data.length; j++) {
			for (i = 0; i < data[0].length; i++) {
				state = data[i][j];
				if (state != 0) {
					setOrientCell(i, j, startX, startY, state, orient, draw);
				}
			}
		}
	}

	public void getFormFromMenu(String string) {
		byte[][] data;
		int ni = getNeighborIndex(), ri;
		Integer ruleIndex = getRuleIndex(ni);

		name = string;
		if (neighbors != LoopRules.indexToNeighbor[ni])
			System.out.println("NeighborIndex: " + ni + ", Neighbors: " +
				neighbors);
		if (ruleIndex == null) {
			System.out.println(string + " ruleIndex is null for neighborIndex " + ni);
			return;
		}
		ri = ruleIndex;
		data = LoopForms.get(ni, ri/*, string*/);
		if (data == null || data.length == 0) {
			System.out.println(string + " data for ruleIndex " + ri + " is null");
			return;
		}
		displayForm(data, true);
	}

	public void getLoop() {
		this.stringSave = ((LoopFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getLoopString();
	}

	public void getLoopFile() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder fileData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				fileData.append(lineData).append(NL);
			this.stringSave = fileData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file to read");
				return;
			}
			getLoopString();
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getLoopString() {
		try {
			String space = " ";
			String newline = NL;
			String pound = "#";
			int dim = 0, neighborhood, cells;
			int width = 5, height = 5;
			int offsetX = 0, offsetY = 0;
			int i = 0;
			String ruleString = "";
			String[] lineToken = this.stringSave.split(newline);
			if (lineToken.length == 0)
				return;
			String lineData = lineToken[i];
			String[] wordToken = null;

			try {
				while (i < lineToken.length) {
					wordToken = lineData.split(space);
					if (wordToken.length > 1 &&
							(pound + "Dim").equals(wordToken[0]))
						break;
					if (++i == lineToken.length)
						return;
					lineData = lineToken[i];
				}
				if (wordToken != null && wordToken.length > 1)
					dim = Integer.parseInt(wordToken[1]);
				if (dim != 2) {
					showMessage(DATAFILE +
						" corrupted: dim " + dim);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: dim not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Neighbors
				neighborhood = Integer.parseInt(wordToken[1]);
				if (Math.abs(neighborhood) <= 2) {
					showMessage(DATAFILE +
						" corrupted: neighbors " + neighborhood);
					neighborhood = this.neighbors;
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: neighbors not an integer");
				neighborhood = this.neighbors;
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #Rule
			//ruleString = LoopRules.parseRule(neighbors, wordToken[1]);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Cells
				cells = Integer.parseInt(wordToken[1]);
				if (cells < 0) {
					showMessage(DATAFILE +
						" corrupted: cells " + cells);
				}
			} catch (Exception e) {
				showMessage(DATAFILE +
					" corrupted: cells not an integer");
			}
			lineData = lineToken[++i];
			// #Name
			name = lineData.substring(lineData.indexOf(space) + 1);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Clockwise
				clockwise = (Integer.parseInt(wordToken[1]) != 0);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: clockwise not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Width
				width = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: width not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Height
				height = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: height not an integer");
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #P
			if (wordToken.length > 1)
				offsetX = Integer.parseInt(wordToken[1]);
			if (wordToken.length > 2)
				offsetY = Integer.parseInt(wordToken[2]);
			//((LoopFrame) frame).callback(
			//	LoopInterface.ACTION_RESTORE);
			cells = scanField(lineToken, i, offsetX, offsetY);
			drawField(true);
			System.out.println(DATAFILE +
				": neighbors " + neighborhood +
				", rule " + ruleString +
				", cells " + cells +
				", name " + name +
				", clockwise " + ((clockwise) ? 1 : 0) +
				", width " + width +
				", height " + height +
				", offsetX " + offsetX +
				", offsetY " + offsetY);
		} catch (NumberFormatException e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading");
		}
	}

	public void writeLoop() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		((LoopFrame) frame).writeTextArea(sb.toString());
	}

	public void writeLoopFile() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.println(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	public void printLoop() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		System.out.println(sb);
	}

	void eraseFrame(Graphics g) {
		g.setColor(background);
		g.fillRect(0, 0, getSize().width, getSize().height);
	}

	void eraseFrame() {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			eraseFrame(g);
		} finally {
			g.dispose();
		}
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		Rectangle r = cellPl.boundRect(ncols, nrows);

		if (g == null || r == null)
			return;
		g.setColor(background.darker().darker());
		g.fillRect(0, 0, getSize().width, r.y);
		g.fillRect(0, r.y, r.x, r.height - r.y);
		g.fillRect(0, r.height, getSize().width, r.y);
		g.fillRect(r.width, r.y,
			getSize().width - r.width, r.height - r.y);
		draw3DFrame(show && focus, r.x, r.y,
			r.width - r.x, r.height - r.y);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void recalibrateBounds(int width, int height) {
		boolean first = true;
		minCol = width / 2 - 1;
		maxCol = width / 2 + 1;
		minRow = height / 2 - 1;
		maxRow = height / 2 + 1;

		for (CellInfo cellInfo : field.list) {
			int position = cellInfo.getPosition();
			int col = getColumn(position, width);
			int row = getRow(position, width);
			if (first) {
				minCol = maxCol = col;
				if (col > 0)
					minCol--;
				if (col < width - 1)
					maxCol++;
				minRow = maxRow = row;
				if (row > 0)
					minRow--;
				if (row < height - 1)
					maxRow++;
				first = false;
			} else {
				if (col - 1 < minCol && col > 0)
					minCol = col - 1;
				if (col + 1 > maxCol && col < width - 1)
					maxCol = col + 1;
				if (row - 1 < minRow && row > 0)
					minRow = row - 1;
				if (row + 1 > maxRow && row < height - 1)
					maxRow = row + 1;
			}
		}
	}

	boolean checkAliveBounds(int i, int j) {
		boolean edge = false;

		if (minCol == 0 || maxCol == ncols - 1 ||
				i == 0 || i == ncols - 1) {
			minCol = 0;
			maxCol = ncols - 1;
			edge = true;
		} else if (i <= minCol) {
			minCol = i - 1;
		} else if (i >= maxCol) {
			maxCol = i + 1;
		}
		if (minRow == 0 || maxRow == nrows - 1 ||
				j == 0 || j == nrows - 1) {
			minRow = 0;
			maxRow = nrows - 1;
			edge = true;
		} else if (j <= minRow) {
			minRow = j - 1;
		} else if (j >= maxRow) {
			maxRow = j + 1;
		}
		return edge;
	}

	int scanField(String[] lineToken, int stringIndex,
			int offsetX, int offsetY) {
		int polygon = getPolygon();
		String lineData;
		int m = stringIndex, count = 0;
		int i, j, n, p, startX, startY, row;
		String space = " ";

		startX = offsetX + ncols / 2;
		startY = offsetY + nrows / 2;
		if (polygon == 3) {
			if (((startX + startY) & 1) == 1) {
				startX--;
			}
		} else if (polygon == 5) {
			startX -= (startX + 2 * startY) % 4;
		}
		try {
			row = 0;
			for (j = startY; ++m < lineToken.length; j++) {
				lineData = lineToken[m];
				if (lineData.length() == 0) {
					break; // allows flexibility
				}
				n = 0;
				i = startX;
				if (polygon == 3) {
					if (((startX + startY) & 1) == 1)
						i--;
				} else if (polygon == 6) {
					i -= (row + (startY & 1)) / 2;
				}
				while (n < lineData.length()) {
					byte oldCell, newCell;
					if (space.charAt(0) == lineData.charAt(n)) {
						n++;
						continue;
					}
					if (i < 0 || i >= ncols ||
							j < 0 || j >= nrows) {
						n++;
						continue;
					}
					p = getPosition(i, j);
					newCell = (byte) (lineData.charAt(n) - '0');
					if (newCell != 0) {
						count++;
					}
					oldCell = field.getCell(p);
					if (oldCell != newCell) {
						field.setCell(p, newCell);
						CellInfo info = new CellInfo(
							p, newCell);
						drawCell(info);
						if (newCell != 0 && oldCell == 0) {
							nCells++;
							checkAliveBounds(i, j);
						} else if (newCell == 0 && oldCell != 0) {
							nCells--;
						}
						((LoopFrame) frame).updateCells(nCells);
					}
					i++;
					n++;
				}
				row++;
			}
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading (scanField)");
		}
		return count;
	}

	public synchronized void drawField(boolean draw) {
		nCells = cellCount();
		for (CellInfo cellInfo : field.list) {
			if (draw)
				drawCell(cellInfo);
		}
		if (draw) {
			if (grid)
				drawGrid();
			((LoopFrame) frame).updateCells(nCells);
			capturer.show();
		}
	}

	void redrawField() {
		eraseFrame();
		drawFrame(true);
		drawField(true);
	}

	public Rectangle getSelectedRectangle() {
		return capturer.getRectangle();
	}

	void banner(String msg) {
		((LoopFrame) frame).displayMessage(msg);
	}

	void drawGrid() {
		Graphics g = this.getGraphics();
		Color color;

		if (g == null)
			return;
		if (grid)
			color = Color.gray;
		else
			color = background;
		g.setColor(color);
		for (int j = 0; j < nrows; j++)
		for (int i = 0; i < ncols; i++) {
				cellPl.drawCell(g, false, null, null, i, j);
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		drawFrame(focus);
		drawField(true);
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}

	int getColumn(int position) {
		return position % this.ncols;
	}

	int getRow(int position) {
		return position / this.ncols;
	}

	static int getColumn(int position, int ncols) {
		return position % ncols;
	}

	static int getRow(int position, int ncols) {
		return position / ncols;
	}

	int getPosition(int col, int row) {
		return getPosition(col, row, this.ncols);
	}

	static int getPosition(int col, int row, int ncols) {
		return row * ncols + col;
	}

	int getColumn(CellInfo info) {
		return getColumn(info.getPosition());
	}

	int getRow(CellInfo info) {
		return getRow(info.getPosition());
	}

	void setPosition(CellInfo info, int col, int row) {
		info.setPosition(getPosition(col, row));
	}

	int[] getNeighborArray(int col, int row) {
		int p;
		int i, n = Math.abs(neighbors);
		int[] c = new int[n];

		if ((col == 0 || row == 0 ||
				col == ncols - 1 || row == nrows - 1) &&
				!wrap) {
			return null;
		}
		for (i = 0; i < n; i++) {
			p = positionOfNeighbor(i, col, row);
			c[i] = field.getCell(p); // non-totalistic
		}
		return c;
	}

	int positionOfNeighbor(int cell, int i, int j) {
		int polygon = getPolygon();
		int col = i, row = j;
		int dir = cell * (360 / Math.abs(neighbors));

		if (neighbors < 0) { /* -6 or -9 */
			dir = (dir >= 180) ? dir - 180 : dir + 180;
		}
		switch (polygon) {
			case 3:
				/* TRI */
				if (((col + row) & 1) == 0) {	/* up */
					switch (dir) {
					case 0: /* close */
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					case 20:
					case 30:
					case 40:
						row = (row + 1 == nrows) ? 0 : row + 1;
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					case 60: /* far */
						row = (row + 1 == nrows) ? 0 : row + 1;
						switch (col) {
						case 0:
							col = ncols - 2;
							break;
						case 1:
							col = ncols - 1;
							break;
						default:
							col = col - 2;
							break;
						}
						break;
					case 80:
					case 90:
					case 100:
						switch (col) {
						case 0:
							col = ncols - 2;
							break;
						case 1:
							col = ncols - 1;
							break;
						default:
							col = col - 2;
							break;
						}
						break;
					case 120: /* close */
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					case 140:
					case 150:
					case 160:
						row = (row == 0) ? nrows - 1 : row - 1;
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					case 180: /* far */
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					case 200:
					case 210:
					case 220:
						row = (row == 0) ? nrows - 1 : row - 1;
						col = (col + 1 == ncols) ? 0 : col + 1;
						break;
					case 240: /* close */
						col = (col + 1 == ncols) ? 0 : col + 1;
						break;
					case 260:
					case 270:
					case 280:
						if (col + 1 == ncols)
							col = 1;
						else if (col + 2 == ncols)
							col = 0;
						else
							col = col + 2;
						break;
					case 300: /* far */
						row = (row + 1 == nrows) ? 0 : row + 1;
						if (col + 1 == ncols)
							col = 1;
						else if (col + 2 == ncols)
							col = 0;
						else
							col = col + 2;
						break;
					case 320:
					case 330:
					case 340:
						row = (row + 1 == nrows) ? 0 : row + 1;
						col = (col + 1 == ncols) ? 0 : col + 1;
						break;
					default:
						System.out.println("wrong direction " + dir);
					}
				} else {	/* down */
					switch (dir) {
					case 0: /* close */
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					case 20:
					case 30:
					case 40:
						row = (row == 0) ? nrows - 1 : row - 1;
						col = (col + 1 == ncols) ? 0 : col + 1;
						break;
					case 60: /* far */
						row = (row == 0) ? nrows - 1 : row - 1;
						if (col + 1 == ncols)
							col = 1;
						else if (col + 2 == ncols)
							col = 0;
						else
							col = col + 2;
							break;
					case 80:
					case 90:
					case 100:
						if (col + 1 == ncols)
							col = 1;
						else if (col + 2 == ncols)
							col = 0;
						else
							col = col + 2;
						break;
					case 120: /* close */
						col = (col + 1 == ncols) ? 0 : col + 1;
						break;
					case 140:
					case 150:
					case 160:
						row = (row + 1 == nrows) ? 0 : row + 1;
						col = (col + 1 == ncols) ? 0 : col + 1;
						break;
					case 180: /* far */
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					case 200:
					case 210:
					case 220:
						row = (row + 1 == nrows) ? 0 : row + 1;
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					case 240: /* close */
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					case 260:
					case 270:
					case 280:
						if (col == 0)
							col = ncols - 2;
						else if ((col - 1) == 0)
							col = ncols - 1;
						else
							col = col - 2;
						break;
					case 300: /* far */
						row = (row == 0) ? nrows - 1 : row - 1;
						if (col == 0)
							col = ncols - 2;
						else if ((col - 1) == 0)
							col = ncols - 1;
						else
							col = col - 2;
						break;
					case 320:
					case 330:
					case 340:
						row = (row == 0) ? nrows - 1 : row - 1;
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					default:
						System.out.println("wrong direction " + dir);
					}
				}
				break;
			case 4:
			case 6:
				switch (dir) {
				case 0:
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 45:
					col = (col + 1 == ncols) ? 0 : col + 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 60:
					if ((row & 1) == 0)
						col = (col + 1 == ncols) ? 0 : col + 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 90:
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 120:
					if ((row & 1) != 0)
						col = (col == 0) ? ncols - 1 : col - 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 135:
					col = (col == 0) ? ncols - 1 : col - 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 180:
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 225:
					col = (col == 0) ? ncols - 1 : col - 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 240:
					if ((row & 1) != 0)
						col = (col == 0) ? ncols - 1 : col - 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 270:
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 300:
					if ((row & 1) == 0)
						col = (col + 1 == ncols) ? 0 : col + 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 315:
					col = (col + 1 == ncols) ? 0 : col + 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			case 5:
				int orient = ((row & 1) * 2 + col) % 4;
				switch (orient) {
				case 0:
					switch (dir) {
					case 0:
						col++;
						break;
					case 51: /* 7 */
					case 72: /* 5 */
						col = (col + 2 >= ncols) ? 0 : col + 2;
						break;
					case 102: /* 7 corner */
						col = (col + 3 >= ncols) ? 1 : col + 3;
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					case 144: /* 5 */
					case 153: /* 7 */
						col++;
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					case 204: /* 7 */
					case 216: /* 5 */
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					case 255: /* 7 */
						col = (col == 0) ? ncols - 1 : col - 1;
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					case 288: /* 5 */
					case 306: /* 7 */
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					default:
						System.out.println("wrong direction " + dir);
					}
					break;
				case 1:
					switch (dir) {
					case 0:
						col--;
						break;
					case 51: /* 7 */
					case 72: /* 5 */
						col = (col == 1) ? ncols - 1 : col - 2;
						break;
					case 102: /* 7 */
						col = (col == 1) ? ncols - 2 : col - 3;
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					case 144: /* 5 */
					case 153: /* 7 */
						col--;
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					case 204: /* 7 */
					case 216: /* 5 */
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					case 255: /* 7 */
						col = (col + 1 >= ncols) ? 0 : col + 1;
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					case 288: /* 5 */
					case 306: /* 7 */
						col = (col + 1 >= ncols) ? 0 : col + 1;
						break;
					default:
						System.out.println("wrong direction " + dir);
					}
					break;
				case 2:
					switch (dir) {
					case 0:
						col = (col + 1 >= ncols) ? 0 : col + 1;
						break;
					case 51: /* 7 */
					case 72: /* 5 */
						row = (row == 0) ? nrows - 1 : row - 1;
						col++;
						break;
					case 102: /* 7 */
						col = (col == 0) ? ncols - 1 : col - 1;
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					case 144: /* 5 */
					case 153: /* 7 */
						col = (col == 0) ? ncols - 2 : col - 2;
						break;
					case 204: /* 7 */
					case 216: /* 5 */
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					case 255: /* 7 */
						row = (row + 1 == nrows) ? 0 : row + 1;
						col = (col == 0) ? ncols - 1 : col - 1;
						break;
					case 288: /* 5 */
					case 306: /* 7 */
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					default:
						System.out.println("wrong direction " + dir);
					}
					break;
				case 3:
					switch (dir) {
					case 0:
						col--;
						break;
					case 51: /* 7 */
					case 72: /* 5 */
						col = (col == 0) ? ncols - 1 : col - 1;
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					case 102: /* 7 */
						col = (col + 1 >= ncols) ? 0 : col + 1;
						row = (row + 1 == nrows) ? 0 : row + 1;
						break;
					case 144: /* 5 */
					case 153: /* 7 */
						col = (col + 2 >= ncols) ? 1 : col + 2;
						break;
					case 204: /* 7 */
					case 216: /* 5 */
						col = (col + 1 >= ncols) ? 0 : col + 1;
						break;
					case 255: /* 7 */
						col = (col + 1 >= ncols) ? 0 : col + 1;
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					case 288: /* 5 */
					case 306: /* 7 */
						row = (row == 0) ? nrows - 1 : row - 1;
						break;
					default:
						System.out.println("wrong direction " + dir);
					}
					break;
				default:
					System.out.println("wrong orient " + orient);
				}
				break;
			default:
				break;
		}
		return getPosition(col, row);
	}

	byte incCell(int col, int row) {
		int p = getPosition(col, row);
		byte state = field.getCell(p);
		byte nextState = (byte) ((state < MAX_STATES - 1) ?
			state + 1 : 0);

		field.setCell(p, nextState);
		if (state < MAX_STATES - 1) {
			nCells++;
			checkAliveBounds(col, row);
		} else {
			nCells--;
		}
		CellInfo info = new CellInfo(p, nextState);
		if (nextState == 0) {
			eraseCell(info);
		} else {
			drawCell(info);
		}
		((LoopFrame) frame).updateCells(nCells);
		return nextState;
	}

	int cellCount() {
		return field.list.size();
	}

	boolean setCell(int col, int row, byte nextState, boolean draw) {
		int p = getPosition(col, row);
		byte state = field.getCell(p);

		if (state == 0 && nextState != 0) {
			field.setCell(p, nextState);
			nCells++;
			checkAliveBounds(col, row);
		} else if (state == nextState) {
			return false;
		} else if (nextState == 0) { // state != 0
			field.setCell(p, (byte) 0);
			nCells--;
		} else { // different sort of live cell
			field.setCell(p, nextState);
		}
		if (draw) {
			CellInfo info = new CellInfo(p, nextState);

			drawCell(info);
			if ((LoopFrame) frame != null)
				((LoopFrame) frame).updateCells(nCells);
		}
		return true;
	}

	/*void printState(int state) {
		CellList curr;
		int i = 0;

		curr = first[state]->next;
		System.out.println("state " + state);
		while (curr != list[state]) {
			System.out.println(i + ": position " +
				curr.info.getPosition() + ",	age " +
				curr.info.getAge() + ", state " +
				curr.info.getCell() + ", toggle " +
				curr.info.getToggle());
			curr = curr.next;
			i++;
		}
	}*/

	static Color colorTable(int state) {
		switch (state) {
		case 0: return Color.black;
		case 1: return Color.blue;
		case 2: return Color.red;
		case 3: return Color.green;
		case 4: return Color.yellow;
		case 5: return Color.magenta;
		case 6: return Color.white;
		case 7: return Color.cyan;
		default: return Color.gray;
		}
	}

	void drawCell(CellInfo info) {
		Graphics g = this.getGraphics();
		Color color;
		int col = getColumn(info);
		int row = getRow(info);
		byte state = info.getCell();

		if (MONO) {
			color = Color.white;
		} else {
			color = colorTable(state);
		}
		cellPl.drawCell(g, false, color, (grid) ? gridColor : background,
			col, row);
	}

	void eraseCell(CellInfo info) {
		Graphics g = this.getGraphics();
		int col = getColumn(info);
		int row = getRow(info);

		cellPl.drawCell(g, false, background,
			(grid) ? gridColor : background, col, row);
	}

	LoopMatrix getField() {
		return field;
	}

class Capturer {
	private Point anchor = new Point(0, 0);
	private Point movable = new Point(0, 0);
	private Point start = new Point(0, 0);
	private Point stop = new Point(0, 0);
	private int completed = -1;
	private final Color selectedColor = Color.magenta;

	Point calcPoint(int x, int y) {
		Point t = new Point(x, y);

		if (t.x < 0)
			t.x = 0;
		if (t.y < 0)
			t.y = 0;
		if (t.x >= ncols)
			t.x = ncols - 1;
		if (t.y >= nrows)
			t.y = nrows - 1;
		return t;
	}

	void calcStartPoint() {
		int polygon = getPolygon();

		if (anchor.x > movable.x)
			start.x = movable.x;
		else
			start.x = anchor.x;
		if (anchor.y > movable.y)
			start.y = movable.y;
		else
			start.y = anchor.y;
		/* System.out.println("calcStartPoint: " +
			start.x + " " + start.y); */
		if (polygon == 3) {
			if (((start.x + start.y) & 1) == 1) {
				if ((start.y & 1) == 1) {
					start.y--;
				} else {
					start.x--;
				}
			}
		} else if (polygon == 5) {
			int mod = (start.x + 2 * start.y) % 4;

			if ((mod / 2) == 1) {
				if ((start.y & 1) == 1) {
					start.y--;
				} else {
					start.x -= 2 ;
				}
			}
			start.x -= (mod & 1);
		} else if (polygon == 6 && (start.y & 1) == 1) {
			start.y--;
		}
	}

	void calcEndPoint() {
		int polygon = getPolygon();

		if (anchor.x < movable.x)
			stop.x = movable.x;
		else
			stop.x = anchor.x;
		if (anchor.y < movable.y)
			stop.y = movable.y;
		else
			stop.y = anchor.y;
		if (polygon == 3) {
			if (((stop.x + stop.y) & 1) == 1) {
				if ((stop.y & 1) == 0)
					stop.y++;
				else
					stop.x++;
			}
		} else if (polygon == 5) {
			int mod = 3 - ((stop.x + 2 * stop.y) % 4);

			if ((mod / 2) == 1) {
				if ((stop.y & 1) == 0) {
					stop.x += 2;
				} else {
					stop.y++;
				}
			}
			stop.x += (mod & 1);
		} else if (polygon == 6 && (stop.y & 1) == 0) {
			stop.y++;
		}
	}

	void startCapture(int column, int row) {
		movable = anchor = calcPoint(column, row);
		calcStartPoint();
		calcEndPoint();
		completed = 0;
		drawCaptureRect((grid) ? gridColor : background);
	}

	void continueCapture(int column, int row) {
		int captureColumns = getColumns();
		int captureRows = getRows();

		movable = calcPoint(column, row);
		drawCaptureRect((grid) ? gridColor : background);
		if (column >= 0 && row >= 0 &&
				column < captureColumns && row < captureRows) {
			calcStartPoint();
			calcEndPoint();
			drawCaptureRect(selectedColor);
			return;
		}
		calcStartPoint();
		calcEndPoint();
	}

	void stopCapture() {
		if (completed == 0)
			completed = 1;
	}

	void clearCapture() {
		drawCaptureRect((grid) ? gridColor : background);
		anchor.setLocation(0, 0);
		movable.setLocation(0, 0);
		completed = -1;
	}

	private void drawCaptureRect(Color color) {
		Graphics g = getGraphics();
		int x = start.x;
		int y = start.y;
		int width = stop.x - start.x + 1;
		int height = stop.y - start.y + 1;

		if (completed == -1)
			return;
		cellPl.drawCaptureRect(g, color,
			x, y, width, height);
	}

	void show() {
		if (!anchor.equals(movable))
			drawCaptureRect(selectedColor);
	}

	Rectangle getRectangle() {
		if (completed == 1) {
			int x = start.x;
			int y = start.y;
			int width = stop.x - start.x + 1;
			int height = stop.y - start.y + 1;

			/*System.out.println("getRectangle: " +
				x + " " + y + " " + width + " " + height); */
			return new Rectangle(x, y, width, height);
		}
		return null;
	}

}

}
