package com.sillycycle.bagleyd.loop;

/*
 * @(#)LoopDialog.java
 *
 * Structure from Java Sourcebook by Ed Anuff, pp 287-289
 *
 * Copyright 2010 - 2018  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.GridLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.Slider;

/**
 * The <code>LoopDialog</code> class holds the configuration
 * setup for the <code>LoopFrame</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class LoopDialog extends Dialog {
	private static final long serialVersionUID = 42L;
	static boolean active = false;
	static LoopFrame frame = null;

	static int columnsValue = LoopInterface.DEFAULT_COLUMNS;
	static int neighborIndexValue =  LoopInterface.DEFAULT_NEIGHBOR_INDEX;
	static boolean randomOrientValue = false;
	static boolean gridValue = true;
	static boolean wrapValue = true;
	static boolean visibleValue = true;

	Slider columnsChanger = new Slider("Field Size",
		columnsValue, LoopInterface.MIN_COLUMNS,
		LoopInterface.MAX_COLUMNS);
	LoopBoxes loopBoxes = new LoopBoxes();
	JCheckBox randomOrientCheckBox = new JCheckBox("Random Orient");
	JCheckBox gridCheckBox = new JCheckBox("Grid");
	JCheckBox wrapCheckBox = new JCheckBox("Wrap");
	JCheckBox visibleCheckBox = new JCheckBox("Visible");

	public LoopDialog(LoopFrame frame, Frame parent,
			String titleText,
			String buttonText) {
		super(parent, titleText, true);

		JButton button = new JButton(buttonText);
		JPanel buttonPanel = new JPanel();
		JPanel sliderPanel = new JPanel();
		JPanel loopBoxesPanel = new JPanel();
		JPanel togglePanel = new JPanel();

		columnsChanger.setWidth(150);
		buttonPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		sliderPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		sliderPanel.add(columnsChanger, BorderLayout.PAGE_START);
		buttonPanel.add(sliderPanel, BorderLayout.PAGE_START);
		loopBoxesPanel.add(loopBoxes, BorderLayout.PAGE_START);
		buttonPanel.add(loopBoxesPanel, BorderLayout.PAGE_START);
		togglePanel.setLayout(new GridLayout(2, 0));
		randomOrientCheckBox.setSelected(randomOrientValue);
		togglePanel.add(randomOrientCheckBox);
		gridCheckBox.setSelected(gridValue);
		togglePanel.add(gridCheckBox);
		wrapCheckBox.setSelected(wrapValue);
		togglePanel.add(wrapCheckBox);
		visibleCheckBox.setSelected(visibleValue);
		togglePanel.add(visibleCheckBox);
		buttonPanel.add(togglePanel, BorderLayout.PAGE_START);

		buttonPanel.add(button, BorderLayout.PAGE_START);
		add(BorderLayout.PAGE_START, buttonPanel);
		pack();
		LoopDialog.active = true;
		LoopDialog.frame = frame;
		columnsChanger.addChangeListener(new ColumnsChangeListener());
		loopBoxes.addActionListener(new LoopActionListener());
		randomOrientCheckBox.addItemListener(new RandomOrientListener());
		gridCheckBox.addItemListener(new GridListener());
		wrapCheckBox.addItemListener(new WrapListener());
		visibleCheckBox.addItemListener(new VisibleListener());

		button.addActionListener((ActionEvent event) -> {
			dispose();
			//else multiple Dialogs can be opened
			active = false;
		});

		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent event) {
				dispose();
				//else multiple Dialogs can be opened
				active = false;
			}
		});
	}

	class ColumnsChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = columnsValue;

			columnsValue = columnsChanger.getValue();
			if (oldValue != columnsValue) {
				frame.endGame();
				frame.setColumns(columnsValue);
				frame.startGame();
			}
		}
	}

	class LoopActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			int ni = neighborIndexValue;

			neighborIndexValue = loopBoxes.getNeighborIndex();
			frame.setRules(ni, neighborIndexValue);
		}
	}

	public static void ruleStateChanged(int ni) {
		frame.setRules(ni, ni);
	}

	class GridListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setGrid(gridCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class RandomOrientListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setRandomOrient(randomOrientCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class WrapListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setWrap(wrapCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class VisibleListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setCellVisible(visibleCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	public void setColumns(int columnsValue) {
		LoopDialog.columnsValue = columnsValue;
		columnsChanger.setValue(columnsValue);
	}

	public static int getColumns() {
		return columnsValue;
	}

	public void setNeighborIndex(int neighborIndexValue) {
		LoopDialog.neighborIndexValue = neighborIndexValue;
		loopBoxes.setNeighborIndex(neighborIndexValue);
	}

	public static int getNeighborIndex() {
		return neighborIndexValue;
	}

	public int getRuleIndex() {
		return loopBoxes.getRuleIndex();
	}

	public void setRandomOrient(boolean flag) {
		LoopDialog.randomOrientValue = flag;
		randomOrientCheckBox.setSelected(flag);
	}

	public static boolean getRandomOrient() {
		return randomOrientValue;
	}

	public void setGrid(boolean flag) {
		LoopDialog.gridValue = flag;
		gridCheckBox.setSelected(flag);
	}

	public static boolean getGrid() {
		return gridValue;
	}

	public void setWrap(boolean flag) {
		LoopDialog.wrapValue = flag;
		wrapCheckBox.setSelected(flag);
	}

	public static boolean getWrap() {
		return wrapValue;
	}

	public void setCellVisible(boolean flag) {
		LoopDialog.visibleValue = flag;
		visibleCheckBox.setSelected(flag);
	}

	public static boolean getCellVisible() {
		return visibleValue;
	}

	public static boolean getActive() {
		return LoopDialog.active;
	}
}
