package com.sillycycle.bagleyd.loop;

/*
 * @(#)LoopForms.java
 *
 * Copyright 2010 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.ArrayList;
import java.util.HashMap;

/**
 * The <code>LoopForms</code> class handles the life form database.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

/*
 * Got some help from
 * http://forums.sun.com/thread.jspa?threadID=530823
 */

/* References:
 */

public class LoopForms {
	enum Category {ADAM, OTHER}
	//private Category kind;
	public LoopForms(int neighborIndex) {
		init(neighborIndex);
	}

	/*-
	Neighborhoods are read as follows (rotations are not listed):
	    N
	  W C E  ==>  I
	    S
	  CNESWI

	   n N
	  w C E  ==>  I
	   s S
	  CNESswnI
	 */

	static final byte[][] selfReproducingLoop = { /* 10x10 */
		{0, 2, 2, 2, 2, 2, 2, 2, 2, 0},
		{2, 4, 0, 1, 4, 0, 1, 1, 1, 2},
		{2, 1, 2, 2, 2, 2, 2, 2, 1, 2},
		{2, 0, 2, 0, 0, 0, 0, 2, 1, 2},
		{2, 7, 2, 0, 0, 0, 0, 2, 7, 2},
		{2, 1, 2, 0, 0, 0, 0, 2, 0, 2},
		{2, 0, 2, 0, 0, 0, 0, 2, 1, 2},
		{2, 7, 2, 2, 2, 2, 2, 2, 7, 2},
		{2, 1, 0, 6, 1, 0, 7, 1, 0, 2},
		{0, 2, 2, 2, 2, 2, 2, 2, 2, 0}
	};

	static final byte[][] bylReproducingLoop = { /* 4x4 */
		{0, 2, 2, 0},
		{2, 3, 1, 2},
		{2, 3, 4, 2},
		{0, 2, 5, 0}
	};

	static final byte[][] chou1ReproducingLoop = { /* 4x4 */
		{0, 0, 0, 0},
		{0, 0, 0, 0},
		{1, 1, 0, 0},
		{3, 4, 1, 1}
	};

	static final byte[][] chou2ReproducingLoop = { /* 3x3 */
		{0, 0, 0},
		{1, 1, 0},
		{3, 4, 1}
	};

	static final byte[][] evoSelfReproducingLoop = { /* 17x17 */
		{0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0},
		{2, 7, 0, 1, 7, 0, 1, 7, 0, 1, 7, 0, 1, 1, 1, 1, 2},
		{2, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 2},
		{2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 2},
		{2, 7, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 2},
		{2, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 2},
		{2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 2},
		{2, 7, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 2},
		{2, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 2},
		{2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 2},
		{2, 7, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 7, 2},
		{2, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2},
		{2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 1, 2},
		{2, 7, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 7, 2},
		{2, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 2},
		{2, 0, 7, 1, 0, 7, 1, 0, 7, 1, 0, 4, 1, 0, 4, 1, 2},
		{0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 5, 0}
	};
	/* TRIA
	              {2,2,0,0,2,7,0},
	             {2,0,2,0,2,0,2},
	            {2,4,1,2,2,1,2},
	           {2,1,2,1,2,1,2},
	          {2,0,2,2,1,1,2},
	         {2,7,1,1,1,1,2},
	        {0,2,2,2,2,2,2},
	              {2,2,0,0,7,7,0},
	             {2,1,2,0,2,0,7},
	            {2,0,1,2,2,1,2},
	           {2,4,2,1,2,1,2},
	          {2,1,2,2,1,1,2},
	         {2,0,7,1,1,1,2},
	        {0,2,2,2,2,2,2}, */

	static final byte[][] hexSelfReproducingLoop = { /* Experimental TRIA5:7x7 */
	/*	      {2,2,0,0,0,0,0},
		     {2,1,2,0,2,2,0},
		    {2,0,4,2,2,0,2},
		   {2,7,2,0,2,0,2},
		  {2,1,2,2,1,1,2},
		 {2,0,7,1,0,7,2},
		{0,2,2,2,2,2,2}, */
	/* Stem cells, only "5" will fully reproduce itself */
	/* 3:12x7 */
	/*	      {2,2,2,2,0,0,0,0,0,0,0,0},
		     {2,1,1,1,2,0,0,0,0,0,0,0},
		    {2,1,2,2,1,2,2,2,2,2,2,0},
		   {2,1,2,0,2,7,1,1,1,1,1,2},
		  {0,2,1,2,2,0,2,2,2,2,2,2},
		 {0,0,2,0,4,1,2,0,0,0,0,0},
		{0,0,0,2,2,2,2,0,0,0,0,0} */
	/* 4:14x9 */
	/*	        {2,2,2,2,2,0,0,0,0,0,0,0,0,0},
		       {2,1,1,1,1,2,0,0,0,0,0,0,0,0},
		      {2,1,2,2,2,1,2,0,0,0,0,0,0,0},
		     {2,1,2,0,0,2,1,2,2,2,2,2,2,0},
		    {2,1,2,0,0,0,2,7,1,1,1,1,1,2},
		   {0,2,1,2,0,0,2,0,2,2,2,2,2,2},
		  {0,0,2,0,2,2,2,1,2,0,0,0,0,0},
		 {0,0,0,2,4,1,0,7,2,0,0,0,0,0},
		{0,0,0,0,2,2,2,2,2,0,0,0,0,0} */
	/* 5:16x11 */
	/*	          {2,2,2,2,2,2,0,0,0,0,0,0,0,0,0,0},
		         {2,1,1,1,1,1,2,0,0,0,0,0,0,0,0,0},
		        {2,1,2,2,2,2,1,2,0,0,0,0,0,0,0,0},
		       {2,1,2,0,0,0,2,1,2,0,0,0,0,0,0,0},
		      {2,1,2,0,0,0,0,2,1,2,2,2,2,2,2,0},
		     {2,1,2,0,0,0,0,0,2,7,1,1,1,1,1,2},
		    {0,2,1,2,0,0,0,0,2,0,2,2,2,2,2,2},
		   {0,0,2,0,2,0,0,0,2,1,2,0,0,0,0,0},
		  {0,0,0,2,4,2,2,2,2,7,2,0,0,0,0,0},
		 {0,0,0,0,2,1,0,7,1,0,2,0,0,0,0,0},
		{0,0,0,0,0,2,2,2,2,2,2,0,0,0,0,0} */
	/* test:3x7  (0,4) is blank  ... very strange.
	          init_adam seems ok something after that I guess */
	/*	             {2,2,0},
		            {2,0,2},
		           {0,2,2},
		          {0,0,0},
		         {2,2,0},
		        {2,1,2},
		       {0,2,2}, */
	/* this might be better for hexagons, space-wise efficient... */
	/* TRIA Experimental TRIA5:7x7 */
	/*	      {2,2,0,0,2,2,0},
		     {2,4,2,0,2,7,2},
		    {2,1,0,2,2,0,2},
		   {2,0,2,1,2,1,2},
		  {2,7,2,2,7,7,2},
		 {2,1,0,7,1,0,2},
		{0,2,2,2,2,2,2}, */
	/* 5:11x11 */
		          {2,2,2,2,2,2,0,0,0,0,0},
		         {2,1,1,7,0,1,2,0,0,0,0},
		        {2,1,2,2,2,2,7,2,0,0,0},
		       {2,1,2,0,0,0,2,0,2,0,0},
		      {2,1,2,0,0,0,0,2,1,2,0},
		     {2,1,2,0,0,0,0,0,2,7,2},
		    {0,2,1,2,0,0,0,0,2,0,2},
		   {0,0,2,1,2,0,0,0,2,1,2},
		  {0,0,0,2,1,2,2,2,2,4,2},
		 {0,0,0,0,2,1,1,1,1,5,2},
		{0,0,0,0,0,2,2,2,2,2,2}
	};

	static final byte[][][][] data = {
	//{ /* 3 Triangular, von Neuman, Faces */ },
	{ /* 4 Square, von Neuman, Faces */
		selfReproducingLoop,
		bylReproducingLoop,
		chou1ReproducingLoop,
		chou2ReproducingLoop,
		evoSelfReproducingLoop,
		selfReproducingLoop,
		selfReproducingLoop,
		selfReproducingLoop,
	},
	//{ /* 5 Pentagonal, Cairo Tiling */ },
	{ /* 6 Hexagonal, von Neuman/Moore */
		hexSelfReproducingLoop,
	},
	};

	static final String[][][]names = {
	//{ /* 3 Triangular, von Neuman, Faces */ },
	{ /* 4 Square, von Neuman, Faces */
		{"selfReproducingLoop"},
		{"bylReproducingLoop"},
		{"chou1ReproducingLoop"},
		{"chou2ReproducingLoop"},
		{"evoSelfReproducingLoop"},
		{"evoSelfReproducingLoop"},
		{"evoSelfReproducingLoop"},
		{"evoSelfReproducingLoop"},
	},
	//{ /* 5 Pentagonal, Cairo Tiling */ },
	{ /* 6 Hexagonal, von Neuman/Moore */
		{"hexSelfReproducingLoop"},
	},
	};

	// Array of Generic Hash Tables
	ArrayList<HashMap<String, Integer>> hashMap = new ArrayList<>();

	final void init(int ni) {
		if (data[ni] == null)
			return;
		for (int ri = 0; ri < data[ni].length; ri++) {
			hashMap.add(new HashMap<String, Integer>());
			int k = 0;
			for (String item : names[ni][ri]) {
				hashMap.get(ri).put(item, k);
				k++;
			}
		}
	}

	//byte[][] get(int neighborIndex, int ruleIndex/*, String s*/) {
//System.out.println("get ni " + neighborIndex + " ruleIndex: " + ruleIndex + " for " + s);
		//return data[neighborIndex][ruleIndex];
	//}

	static byte[][] get(int neighborIndex, int ruleIndex) {
		return data[neighborIndex][ruleIndex];
	}

	static Category getCategoryIndex(int neighborIndex, int ruleIndex,
			int formIndex) {

		if (formIndex < names[neighborIndex][ruleIndex].length)
			return Category.ADAM;
		return Category.OTHER; // not really correct but ok
	}

	static String getName(int neighborIndex, int ruleIndex, int formIndex) {
		if (formIndex < names[neighborIndex][ruleIndex].length)
			return names[neighborIndex][ruleIndex][formIndex];
		return null;
	}

	static int length(int neighborIndex, int ruleIndex) {
		return data[neighborIndex][ruleIndex].length;
	}
}
