package com.sillycycle.bagleyd.loop;

/*
 * @(#)LoopFrame.java
 *
 * Copyright 2010 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.IODialog;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;
import com.sillycycle.bagleyd.util.TrafficLight;

/**
 * The <code>LoopFrame</code> class is to be embedded in a Web page
 * or run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class LoopFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final String[] COMMAND_STRINGS = new String[] {
		"Get Buffer", "Write Buffer", "Show Buffer",
		"Print", "Forms",
		"Stop", "Step (Yield)", "Run", "Clear",
		"Speed >", "Slow <", "Setup ~",
		"Description ?", "Features !", "References ^", "About",
		"Get File", "Write File", "Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'g', 'w', '*', 'p', 'f',
		's', 'y', 'r', 'c', '>', '<', '~',
		'?', '!', '^', 'a', 'G', 'W', 'x'};
	static final String[] TRAFFIC_STRINGS = new String[] {
		"Stop", "Step", "Run", "Clear"
	};
	static final char[] TRAFFIC_CHARS = new char[] {
		's', 'y', 'r', 'c'
	};
	static final Color[] TRAFFIC_COLORS = new Color[] {
		Color.red, Color.yellow, Color.green, Color.blue // For Jimmi
	};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 806;
	int heightValue = 456;
	static int xValue = 0;
	static int yValue = 0;
	Color bgColor = Color.gray.darker();
	boolean controlValue = true;
	int speedValue = LoopInterface.DEFAULT_SPEED;
	boolean debug = false;
	Loop game = null;
	JPanel topPanel = new JPanel();
	TrafficLight trafficLight = new TrafficLight(TRAFFIC_STRINGS,
		TRAFFIC_CHARS, TRAFFIC_COLORS);
	JTextField ruleTextField = new JTextField(15);
	JTextField genTextField = new JTextField(5);
	JTextField cellsTextField = new JTextField(5);
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	LoopDialog gameDialog = null;
	IODialog ioDialog = null;
	static int neighborIndex = LoopInterface.DEFAULT_NEIGHBOR_INDEX;
	JMenuBar menuBar;
	JMenuItem detachMenuItem;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	Spinner speedChanger = null;
	//Slider speedChanger = null;

	public String writeLoopString() {
		return this.stringBuilderSave.toString();
	}

	public void getLoopString(String s) {
		game.stringSave = s;
	}

	public void startGame() {
		game.start();
	}

	public void endGame() {
		game.setPause(true);
	}

	public void pauseGame() {
		game.setPause(!game.getPause());
	}

	public void clearGame() {
		game.clearField(true);
		game.start();
	}

	public void processKey(char key) {
		game.processKey(key);
	}

	public synchronized void setDelay(int i) {
		this.speedValue = i;
		if (speedChanger != null)
			speedChanger.setValue(LoopInterface.MAX_SPEED +
				LoopInterface.MIN_SPEED - 1 - i);
	}

	public void setColumns(int i) {
		if (game == null) {
			gameDialog.setColumns(i);
		} else {
			game.setColumns(i);
			gameDialog.setColumns(game.getColumns());
		}
	}

	public static int getColumns() {
		return LoopDialog.getColumns();
	}

	public void setNeighborIndex(int i) {
		gameDialog.setNeighborIndex(i);
		if (game != null)
			game.setNeighborIndex(i);
	}

	public static int getNeighborIndex() {
		return LoopDialog.getNeighborIndex();
	}

	public Integer getRuleIndex() {
		//return game.getRuleIndex(neighborIndex);
		// directly as there is no survival or birth stuff
		return gameDialog.getRuleIndex();
	}

	public void setRules(int oldNeighborIndex, int newNeighborIndex) {
		if (game != null) {
			game.setNeighbors(newNeighborIndex);
			game.ruleChange();
			if (LoopCanvas.polygonChange(oldNeighborIndex, newNeighborIndex)
) {
				game.eraseFrame();
				game.initTable();
				game.clearField(true);
				game.resizeField();
			} else
				game.clearField(true);
		}
	}

	public void setRule(/*int neighborIndex, */String rule) {
		updateRule(rule);
	}

	public void setGrid(boolean flag) {
		gameDialog.setGrid(flag);
		if (game != null)
			game.setGrid(flag);
	}

	public static boolean getGrid() {
		return LoopDialog.getGrid();
	}

	/*public void setInvert(boolean flag) {
		gameDialog.setInvert(flag);
		if (game != null)
			game.setInvert(flag);
	}

	public boolean getInvert() {
		return gameDialog.getInvert();
	}*/

	public void setRandomOrient(boolean flag) {
		gameDialog.setRandomOrient(flag);
		if (game != null)
			game.setRandomOrient(flag);
	}

	public static boolean getRandomOrient() {
		return LoopDialog.getRandomOrient();
	}

	public void setWrap(boolean flag) {
		gameDialog.setWrap(flag);
		if (game != null)
			game.setWrap(flag);
	}

	public static boolean getWrap() {
		return LoopDialog.getWrap();
	}

	public void setCellVisible(boolean flag) {
		gameDialog.setCellVisible(flag);
		if (game != null)
			game.setCellVisible(flag);
	}

	public static boolean getCellVisible() {
		return LoopDialog.getCellVisible();
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		int oldValue, newValue;
		int neighborsValue;
		int delayValue;
		int oldNeighborIndex;

		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		bgColor = getOpt("bgColor", bgColor);
		delayValue = LoopInterface.MAX_SPEED +
			LoopInterface.MIN_SPEED - 1 - speedValue;
		delayValue = getOpt("delay", delayValue);
		speedValue = LoopInterface.MAX_SPEED +
			LoopInterface.MIN_SPEED - 1 - delayValue;
		oldValue = LoopDialog.getColumns();
		newValue = getOpt("columns", oldValue);
		if (newValue < LoopInterface.MIN_COLUMNS)
			newValue = LoopInterface.MIN_COLUMNS;
		if (oldValue != newValue)
			gameDialog.setColumns(newValue);
		neighborsValue = getOpt("neighbors",
			LoopRules.indexToNeighbor[LoopDialog.getNeighborIndex()]);
		oldNeighborIndex = neighborIndex;
		neighborIndex = LoopRules.neighborToIndex(neighborsValue);
		if (oldNeighborIndex != neighborIndex) {
			gameDialog.setNeighborIndex(neighborIndex);
		}
		gameDialog.setRandomOrient(getOpt("randomOrient",
			LoopDialog.getRandomOrient()));
		gameDialog.setGrid(getOpt("grid",
			LoopDialog.getGrid()));
		gameDialog.setWrap(getOpt("wrap",
			LoopDialog.getWrap()));
		gameDialog.setCellVisible(getOpt("visible",
			LoopDialog.getCellVisible()));
	}

	public void getOpts2() {
		String ruleString = getOpt("rule");

		if (ruleString != null) {
			String newRule;
			String neighborRule;
			/*int neighborIndex = gameDialog.getNeighborIndex();*/

			newRule = LoopRules.parse(
				Integer.toString(LoopRules.indexToNeighbor[LoopDialog.getNeighborIndex()]) +
				ruleString);
			neighborRule = LoopRules.parseNeighborRule(newRule);
			setRule(/*neighborIndex, */neighborRule);
		}
	}

	public void setupGame() {
		if (gameDialog != null) {
			gameDialog.setVisible(true);
		}
	}

	public void showBuffer() {
		if (ioDialog != null) {
			ioDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Loop", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 350, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"Yet another implementation of Langton's Loops.  This one allows not" + NL +
"  only square life but also hexagonal tiling." + NL +
"When running from web the \"write\" and \"read\" use cookies to output " + NL +
"data, else it uses \"" + LoopCanvas.DATAFILE + "\" file." + NL +
"Holding Control key down and use mouse to copy and paste." + NL +
"Can also manipulate life form by rotating and reflecting." + NL +
"See https://www.sillycycle.com/lifeNotes.html for TODO list. " + NL +
"Probably lots of other things need to be done too, let me know." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"Press mouse button on a on cell to erase (drag to do multiple)." + NL +
" " + NL +
"Press mouse button on a off cell to turn it on (drag to do multiple)." + NL +
" " + NL +
"Press \"g\" key to get (read) a saved life form to screen." + NL +
" " + NL +
"Press \"w\" key to save (write) a life form to screen." + NL +
" " + NL +
"Press \"G\" key to get (read) a saved life form to file." + NL +
" " + NL +
"Press \"W\" key to save (write) a life form to file." + NL +
" " + NL +
"Press \"P\" or \"p\" keys to print a life form." + NL +
" " + NL +
"Press \"C\" or \"c\" keys to clear." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to wrap." + NL +
" " + NL +
"Press \"#\" key to toggle grid." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up loop." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow down loop." + NL +
" " + NL +
"Press \"]\" or \"}\" keys to increase the column/row " +
"size." + NL +
" " + NL +
"Press \"[\" or \"{\" keys to decrease the column/row " +
"size." + NL +
" " + NL +
"Press \"Y\" or \"y\" keys to step." + NL +
" " + NL +
"Press \"R\", \"r\", or \"|\" key to toggle run." + NL +
" " + NL +
"Press \"S\" or \"s\" key to toggle stop run." + NL +
" " + NL +
"Press \"~\" key for setup." + NL +
" " + NL +
"Press \"Esc\" key to hide and pause program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program if an application." + NL,
/* "The loop may be resized.  Cells will reshape depending on" + NL +
"the room they have." + NL + */
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"xlock loop source code" + NL +
"https://www.sillycycle.com/xlockmore.html" + NL +
" " + NL +
"Langton, Chris, \"Self-Reproduction in Cellular Automata\"" + NL +
"Physica10D 1984, pages 135-144." + NL +
" " + NL +
"From Steven Levy's Artificial Life" + NL +
"Chris Langton's cellular automata \"loops\" reproduce in the spirit of life." + NL +
"Beginning from a single organism, the loops from a colony.  As the loops" + NL +
"on the outer fringes reproduce, the inner loops -- blocked by their" + NL +
"daughters -- can no longer produce offspring.  These dead progenitors" + NL +
"provide a base for future generations' expansion, much like the formation" + NL +
"of a coral reef.  This self-organizing behavior emerges spontaneously," + NL +
"from the bottom up -- a key characteristic of artificial life.",
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java LoopFrame Version 8.3.0" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/life.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = LoopFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	@Override
	public void setSize(int width, int height) {
		if (game == null)
			return;
		if (width != widthValue || height != heightValue) {
			game.setBounds(0, 0, width, height);
			game.resizeField();
			widthValue = width;
			heightValue = height;
			game.drawField(true);
		}
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		setSize(secondaryPanel.getSize().width,
			secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public String readTextArea() {
		return ioDialog.getText();
	}

	public void writeTextArea(String s) {
		ioDialog.setText(s);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	public void displayMessage(String string) {
		if (frame != null) {
			if (string == null || "".equals(string))
				frame.setTitle(LoopInterface.TITLE);
			else
				frame.setTitle(LoopInterface.TITLE +
					": " + string);
		}
	}

	@Override
	public void requestFocus() {
		if (game != null)
			game.requestFocus();
	}

	public void updateRule(String string) {
		if (ruleTextField == null)
			return;
		ruleTextField.setText(string);
	}

	public void updateGen(int gen) {
		if (genTextField == null)
			return;
		genTextField.setText(String.valueOf(gen));
	}

	public void updateCells(int cells) {
		if (cellsTextField == null)
			return;
		cellsTextField.setText(String.valueOf(cells));
	}

	public void initPopup() {
		if (game != null)
			game.initPopup();
	}

	public Image getImage(String fileName) {
		return Toolkit.getDefaultToolkit().getImage(fileName);
	}

	class PressActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			processKey(event.getActionCommand().charAt(0));
		}
	}

	class SpeedChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = speedValue;

			speedValue = speedChanger.getValue();
			if (oldValue != speedValue) {
				game.setDelay(LoopInterface.MAX_SPEED +
					LoopInterface.MIN_SPEED - 1 -
					speedValue);
			}
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt]);
				return;
			}
		}
	}

	public LoopFrame(String title) {
		JPanel controlPanel = new JPanel();
		JPanel textPanel = new JPanel();
		JPanel genPanel = new JPanel();
		JPanel canvasPanel = new JPanel();

		try {
			mediumIcon = getIcon(LoopInterface.ICONS[LoopInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(LoopInterface.ICONS[LoopInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		gameDialog = new LoopDialog(this,
			ComponentUtil.findJFrame(this),
			LoopInterface.TITLE + ": Setup", "DONE");
		ioDialog = new IODialog(ComponentUtil.findJFrame(this),
			LoopInterface.TITLE + ": IO", "Life Form", "DONE");
		getOpts();
		topPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		controlPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 3, 3));
		textPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 3, 3));
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Get Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_G);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_W);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Show Buffer");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Print");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_P);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Get File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//LoopFrame.formsMenu = new JMenu("Forms");
			//LoopFrame.formsMenuItems = LoopFrame.updateForms(
			//	LoopFrame.formsMenu,
			//	LoopFrame.formsMenuItems,
			//	frame.getNeighborIndex(),
			//	frame.getRuleIndex(), this);
			//menuBar.add(LoopFrame.formsMenu);
			//menuItem.addActionListener(this);
			//menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Stop");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_S);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Step (Yield)");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Y);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Run");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_R);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Speed >");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Slow <");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Setup ~");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			controlPanel.add(trafficLight);
			genPanel.add(new JLabel("Rule:"));
			genPanel.add(ruleTextField);
			ruleTextField.setEditable(false);
			genPanel.add(new JLabel("Generation:"));
			genPanel.add(genTextField);
			genTextField.setEditable(false);
			genPanel.add(new JLabel("Cells:"));
			genPanel.add(cellsTextField);
			cellsTextField.setEditable(false);
			speedChanger = new Spinner("Speed",
				speedValue, LoopInterface.MIN_SPEED,
				LoopInterface.MAX_SPEED);
			/*speedChanger = new Slider("Speed",
				speedValue, LoopInterface.MIN_SPEED,
				LoopInterface.MAX_SPEED);
			speedChanger.setWidth(120);*/
			textPanel.add(genPanel);
			controlPanel.add(speedChanger);
			gameDialog.setNeighborIndex(neighborIndex);
			ruleTextField.setText(//LoopRules.indexToNeighbor[neighborIndex] +
				LoopRules.defaultRuleStrings[neighborIndex]);
			genTextField.setText("0");
			cellsTextField.setText("0");
			topPanel.add(controlPanel);
			topPanel.add(textPanel);
			secondaryPanel.setLayout(new BorderLayout());
			secondaryPanel.add(topPanel, BorderLayout.PAGE_START);
		}
		game = new Loop(this, speedValue,
			LoopDialog.getColumns(),
			LoopRules.indexToNeighbor[LoopDialog.getNeighborIndex()],
			LoopDialog.getRandomOrient(),
			LoopDialog.getGrid(), LoopDialog.getWrap(),
			LoopDialog.getCellVisible(), bgColor);
		canvasPanel.add(game);
		secondaryPanel.add(canvasPanel, BorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			trafficLight.addActionListener(new PressActionListener());
			speedChanger.addChangeListener(new SpeedChangeListener());
			ruleTextField.addActionListener((ActionEvent event) -> {
				String ruleBuffer = ruleTextField.getText();
				String newRule = LoopRules.parse(ruleBuffer);
				String neighborRule = LoopRules.parseNeighborRule(newRule);
				int newNeighborIndex = LoopRules.parseNeighborFromRule(newRule);

//				System.out.println("init: neighborIndex " +
//					neighborIndex + ", newNeighborIndex + " + newNeighborIndex);

				if (neighborIndex != newNeighborIndex) {
					setRules(neighborIndex, newNeighborIndex);
					game.drawField(true);
					neighborIndex = newNeighborIndex;
					gameDialog.setNeighborIndex(neighborIndex);
				}
				setRule(/*newNeighborIndex, */neighborRule);
			});
		}
		getOpts2();
		game.setBounds(0, 0, widthValue, heightValue);
		//show();
		setTitle(title);
		game.requestFocus();
		game.setPause(true);
		/* game.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		//game.setSize(widthValue, heightValue);
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : LoopInterface.ICONS) {
			URL url = LoopFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-bgColor={int}] [-control={bool}]" + NL +
"\t[-columns={int}] [-neighbors={int}] [-rule={string}]" + NL +
"\t[-randomOrient={bool}] [-grid={bool}] [-wrap={bool}]" + NL +
"\t[-visible={bool}] [-delay={int}]");
			return;
		}
		LoopFrame frame = new LoopFrame(LoopInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.game.requestFocusInWindow();
		frame.setLocation(LoopFrame.xValue, LoopFrame.yValue);
		frame.setVisible(true);
	}
}
