package com.sillycycle.bagleyd.loop;

/* Code for selection spliced in is due to Alex Shvets */
/* http://home.comcast.net/~shvets/Programs/Examples/GameLife/life2.html */
// LoopPopup.java

import java.awt.Cursor;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;

public class LoopPopup implements ClipboardOwner, ActionListener {
	Clipboard clipboard = new Clipboard("Life");
	LoopCanvas loop;
	static final boolean DEBUG = false;

	public LoopPopup(LoopCanvas loop) {
		this.loop = loop;
	}

	@Override
	public void lostOwnership(Clipboard lostClipboard, Transferable contents) {
		// implements ClipboardOwner
	}

	// implements ActionListener
	@Override
	public void actionPerformed(ActionEvent e) {
		String command = e.getActionCommand();
		Rectangle selectedRect = getSelectedRectangle();
		LoopMatrix field = loop.getField();

		if (DEBUG) {
			if (selectedRect == null)
				System.out.println("actionPeformed: selectedRect = null");
			else if (loop.getField() == null)
				System.out.println("actionPeformed: field = null");
		}
		if (selectedRect == null || field == null)
			return;
		if (DEBUG) {
			System.out.println("actionPeformed: " + command);
		}
		loop.setCursor(new Cursor(Cursor.WAIT_CURSOR));
		switch(command) {
		case "Cut":
			doCut(selectedRect);
			break;
		case "Copy":
			doCopy(selectedRect);
			break;
		case "Paste":
			doPaste(selectedRect);
			break;
		case "Clear":
			doClear(selectedRect);
			break;
		case "Cw":
			doRotate(selectedRect, true);
			break;
		case "Ccw":
			doRotate(selectedRect, false);
			break;
		case "Refl Wall":
			doReflect(selectedRect, true);
			break;
		case "Refl Grnd":
			doReflect(selectedRect, false);
			break;
		default:
		}
		loop.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	}

	public Rectangle getSelectedRectangle() {
		return loop.getSelectedRectangle();
	}

	public void doClear(Rectangle selectedRect) {
		LoopMatrix field = loop.getField();

		field.clearSubMatrix(selectedRect);
		loop.redrawField(); //(selectedRect);
	}

	public void doCopy(Rectangle selectedRect) {
		LoopMatrix field = loop.getField();

		LoopMatrix selection = field.getSubMatrix(selectedRect);
		LoopSelection data = new LoopSelection(selection);
		clipboard.setContents(data, this);
	}

	public void doCut(Rectangle selectedRect) {
		LoopMatrix field = loop.getField();
		LoopMatrix selection = field.getSubMatrix(selectedRect);
		LoopSelection data = new LoopSelection(selection);

		clipboard.setContents(data, this);

		field.clearSubMatrix(selectedRect);
		loop.redrawField(); //(selectedRect);
	}

	public void doPaste(Rectangle selectedRect) {
		LoopMatrix field = loop.getField();

		try {
			Transferable data = clipboard.getContents(this);
			if (data == null) {
				return;
			}
			LoopMatrix selection = (LoopMatrix) data.getTransferData(
				LoopSelection.lifeFlavor);

			selectedRect.setSize(selection.getColumns(),
				selection.getRows());
			field.setSubMatrix(selection, selectedRect);
			loop.redrawField(); //(selectedRect);
		} catch (UnsupportedFlavorException | IOException e) {
			e.printStackTrace();
		}
	}

	public void doRotate(Rectangle selectedRect, boolean cw) {
		LoopMatrix field = loop.getField();

		try {
			Transferable data = clipboard.getContents(this);
			if (data == null) {
				return;
			}
			LoopMatrix selection = (LoopMatrix) data.getTransferData(
				LoopSelection.lifeFlavor);

			selectedRect.setSize(selection.getColumns(),
				selection.getRows());
			field.rotate(selection, selectedRect,
				loop.getPolygon(), cw);
			loop.redrawField(); //(selectedRect);
		} catch (UnsupportedFlavorException | IOException e) {
			e.printStackTrace();
		}
	}

	public void doReflect(Rectangle selectedRect, boolean wall) {
		LoopMatrix field = loop.getField();

		try {
			Transferable data = clipboard.getContents(this);
			if (data == null) {
				return;
			}
			LoopMatrix selection = (LoopMatrix) data.getTransferData(
				LoopSelection.lifeFlavor);

			selectedRect.setSize(selection.getColumns(),
				selection.getRows());
			field.reflect(selection, selectedRect,
				loop.getPolygon(), wall);
			loop.redrawField(); //(selectedRect);
		} catch (UnsupportedFlavorException | IOException e) {
			e.printStackTrace();
		}
	}
}
