package com.sillycycle.bagleyd.loop;

/*
 * @(#)LoopRules.java
 *
 * Copyright 2010 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.ArrayList;
import java.util.HashMap;

/**
 * The <code>LoopRules</code> class handles the rules of Loop's Loops.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class LoopRules {
	static final int MAX_NEIGHBORS = 6;
	static final int MAX_NEIGHBOR_INDEXES = 2;
	static final int BASE = 10;
	static boolean dissolve = false;
	static boolean sheath = false;
	static final boolean DEBUG = false;

	private String discoverer;
	private int neighborIndex;
	private String ruleName;
	boolean langton = true;
	boolean evolve = false;

	public LoopRules(String ruleName, String discoverer) {
		this.discoverer = discoverer;
		this.ruleName = ruleName;
	}

	public LoopRules(int neighbors, String ruleName) {
		this.discoverer = "";
		this.neighborIndex = neighborToIndex(neighbors);
		this.ruleName = ruleName;
		init();
	}

	public LoopRules(int neighbors, boolean langton, boolean evolve) {
		this.neighborIndex = neighborToIndex(neighbors);

		this.langton = langton;
		this.evolve = evolve;
		discoverer = "";
		/*for (int i = 0; i < loopRules[neighborIndex].length; i++) {
			discoverer = loopRules[neighborIndex][i].discoverer;
			break;
		}*/
		init();
	}

	protected static String toString(int[] values) {
		StringBuilder string = new StringBuilder();

		for (int i = 0; i < values.length; i++) {
			if (i != 0)
				string.append(", ");
			string.append(i).append(": ").append(values[i]);
		}
		return string.toString();
	}

	public String getDiscoverer() {
		return discoverer;
	}

	public int getNeighbors() {
		return indexToNeighbor[neighborIndex];
	}

	public String getRuleName() {
		return ruleName;
	}

	public static String getRuleName(int neighborIndexValue, int ruleIndexValue) {
//		System.out.println("getRuleName: " + neighborIndexValue + ", " + ruleIndexValue);
		return loopRules[neighborIndexValue][ruleIndexValue].ruleName;
	}

	public void setDiscoverer(String discoverer) {
		this.discoverer = discoverer;
	}

	public void setNeighborIndex(int neighborIndex) {
		this.neighborIndex = neighborIndex;
	}

	public void setNeighbors(int neighbors) {
		this.neighborIndex = neighborToIndex(neighbors);
	}

	public void setRuleName(String ruleName) {
		this.ruleName = ruleName;
	}

	/* C.G.Langton. "Self-reproduction in cellular automata." */
	/* Physica D, Vol. 10, pages 135-144, 1984. */
	/* transition rules from: http://necsi.org/postdocs/sayama/sdsr/java/loops.java */
	/* credits: "Self-Replicating Loops & Ant, Programmed by Eli Bachmutsky, Copyleft Feb.1999" */
	/* states: 8, rules: 219, format: CNESWC' */
	static int[] transitionTable = { /* Octal  CNESW->I */
	      /* CNESWI   CNESWI   CNESWI   CNESWI   CNESWI */
		0000000, 0000012, 0000020, 0000030, 0000050,
		0000063, 0000071, 0000112, 0000122, 0000132,
		0000212, 0000220, 0000230, 0000262, 0000272,
		0000320, 0000525, 0000622, 0000722, 0001022,
		0001120, 0002020, 0002030, 0002050, 0002125,
		0002220, 0002322, 0005222, 0012321, 0012421,
		0012525, 0012621, 0012721, 0012751, 0014221,
		0014321, 0014421, 0014721, 0016251, 0017221,
		0017255, 0017521, 0017621, 0017721, 0025271,
		0100011, 0100061, 0100077, 0100111, 0100121,
		0100211, 0100244, 0100277, 0100511, 0101011,
		0101111, 0101244, 0101277, 0102026, 0102121,
		0102211, 0102244, 0102263, 0102277, 0102327,
		0102424, 0102626, 0102644, 0102677, 0102710,
		0102727, 0105427, 0111121, 0111221, 0111244,
		0111251, 0111261, 0111277, 0111522, 0112121,
		0112221, 0112244, 0112251, 0112277, 0112321,
		0112424, 0112621, 0112727, 0113221, 0122244,
		0122277, 0122434, 0122547, 0123244, 0123277,
		0124255, 0124267, 0125275, 0200012, 0200022,
		0200042, 0200071, 0200122, 0200152, 0200212,
		0200222, 0200232, 0200242, 0200250, 0200262,
		0200272, 0200326, 0200423, 0200517, 0200522,
		0200575, 0200722, 0201022, 0201122, 0201222,
		0201422, 0201722, 0202022, 0202032, 0202052,
		0202073, 0202122, 0202152, 0202212, 0202222,
		0202272, 0202321, 0202422, 0202452, 0202520,
		0202552, 0202622, 0202722, 0203122, 0203216,
		0203226, 0203422, 0204222, 0205122, 0205212,
		0205222, 0205521, 0205725, 0206222, 0206722,
		0207122, 0207222, 0207422, 0207722, 0211222,
		0211261, 0212222, 0212242, 0212262, 0212272,
		0214222, 0215222, 0216222, 0217222, 0222272,
		0222442, 0222462, 0222762, 0222772, 0300013,
		0300022, 0300041, 0300076, 0300123, 0300421,
		0300622, 0301021, 0301220, 0302511, 0401120,
		0401220, 0401250, 0402120, 0402221, 0402326,
		0402520, 0403221, 0500022, 0500215, 0500225,
		0500232, 0500272, 0500520, 0502022, 0502122,
		0502152, 0502220, 0502244, 0502722, 0512122,
		0512220, 0512422, 0512722, 0600011, 0600021,
		0602120, 0612125, 0612131, 0612225, 0700077,
		0701120, 0701220, 0701250, 0702120, 0702221,
		0702251, 0702321, 0702525, 0702720
	};

	/* J. Byl. "Self-Reproduction in small cellular automata." */
	/* Physica D, Vol. 34, pages 295-299, 1989. */
	/* transition rules from: http://necsi.org/postdocs/sayama/sdsr/java/loops.java */
	/* credits: "Self-Replicating Loops & Ant, Programmed by Eli Bachmutsky, Copyleft Feb.1999" */
	/* states: 6, rules: 146, format: CNESWC' */
	static int[] bylTransitionTable = {
		0000000, 0000010, 0000031, 0000420, 0000110,
		0000120, 0000311, 0000330, 0000320, 0000200,
		0000242, 0000233, 0004330, 0005120, 0001400,
		0001120, 0001305, 0001310, 0001320, 0003020,
		0003400, 0003420, 0003120, 0002040, 0002050,
		0002010, 0002020, 0002420, 0002500, 0002520,
		0002200, 0002250, 0002220, 0040000, 0050000,
		0050500, 0010052, 0010022, 0010400, 0010100,
		0020055, 0400233, 0405235, 0401233, 0442024,
		0452020, 0415233, 0411233, 0412024, 0412533,
		0432024, 0421433, 0422313, 0501302, 0502230,
		0540022, 0542002, 0512024, 0530025, 0520025,
		0520442, 0523242, 0522020, 0100000, 0100010,
		0100033, 0100330, 0101233, 0103401, 0103244,
		0111244, 0113244, 0112404, 0133244, 0121351,
		0123444, 0123543, 0122414, 0122434, 0300100,
		0300300, 0300211, 0300233, 0304233, 0301100,
		0301211, 0303211, 0303233, 0302233, 0344233,
		0343233, 0351202, 0353215, 0314233, 0311233,
		0313251, 0313211, 0312211, 0335223, 0333211,
		0320533, 0324433, 0325415, 0321433, 0321511,
		0321321, 0323411, 0200000, 0200042, 0200032,
		0200022, 0200442, 0200515, 0200112, 0200122,
		0200342, 0200332, 0200242, 0200212, 0201502,
		0203202, 0202302, 0244022, 0245022, 0242042,
		0242022, 0254202, 0255042, 0252025, 0210042,
		0214022, 0215022, /*0212052,*/ 0212055, 0212022,
		0230052, 0234022, 0235002, 0235022, 0232042,
		0232022, 0220042, 0220020, 0220533, 0221552,
		0322343, 0204122,
 /* last 2 are not listed but implied by initializing rules to be stable */
	};

	/* J. A. Reggia, S.L.Armentrout, H.-H. Chou, and Y. Peng. */
	/* Simple systems that exhibit self-directed replication. */
	/* Science, Vol. 259, pages 1282-1287, February 1993. */
	/* transition rules from: http://necsi.org/postdocs/sayama/sdsr/java/loops.java */
	/* credits: "Self-Replicating Loops & Ant, Programmed by Eli Bachmutsky, Copyleft Feb.1999" */
	/* rules: 174, states: 8, format: CNESWC' */
	static int[] chou1TransitionTable = {
		0000000, 0000200, 0000110, 0004420, 0006301,
		0001447, 0003030, 0046000, 0060400, 0062102,
		0051060, 0010420, 0010360, 0012152, 0032000,
		0401033, 0410011, 0431206, 0640406, 0660630,
		0616600, 0636000, 0000440, 0000260, 0000330,
		0004512, 0005103, 0001540, 0003100, 0045002,
		0060600, 0061110, 0051710, 0010506, 0010320,
		0030400, 0076010, 0460313, 0410633, 0600000,
		0640106, 0610006, 0616100, 0636100, 0000420,
		0000231, 0000705, 0006030, 0002020, 0001210,
		0040000, 0045600, 0060100, 0050006, 0020320,
		0010140, 0016000, 0036000, 0400035, 0466633,
		0410100, 0606100, 0660060, 0610106, 0613001,
		0676600, 0000600, 0000100, 0004040, 0006440,
		0001010, 0001110, 0040160, 0041040, 0060166,
		0050306, 0024200, 0010154, 0016100, 0036606,
		0400313, 0423106, 0411033, 0606313, 0660460,
		0614004, 0630160, 0500000, 0000540, 0000152,
		0004010, 0006110, 0001030, 0003000, 0044400,
		0060066, 0062030, 0054040, 0023010, 0010340,
		0012000, 0036300, 0405033, 0410061, 0431033,
		0640006, 0660615, 0616000, 0630360, 0500044,
		0500011, 0230000, 0100111, 0101044, 0146004,
		0143321, 0166611, 0121121, 0117044, 0300111,
		0340066, 0366611, 0374017, 0200100, 0100044,
		0104414, 0101601, 0141403, 0160414, 0161101,
		0112244, 0130414, 0304011, 0340611, 0314001,
		0703010, 0203010, 0100011, 0104674, 0101301,
		0141424, 0160111, 0161301, 0111044, 0133001,
		0301471, 0340261, 0314676, 0703330, 0240400,
		0100441, 0106101, 0103011, 0141141, 0164104,
		0120414, 0113011, 0133011, 0301303, 0340211,
		0313023, 0733630, 0210102, 0100414, 0105011,
		0103103, 0141121, 0166644, 0124104, 0113601,
		0300101, 0307141, 0344011, 0331001,
	};

	/* J. A. Reggia, S.L.Armentrout, H.-H. Chou, and Y. Peng. */
	/* Simple systems that exhibit self-directed replication. */
	/* Science, Vol. 259, pages 1282-1287, February 1993. */
	/* transition rules from: http://necsi.org/postdocs/sayama/sdsr/java/loops.java */
	/* credits: "Self-Replicating Loops & Ant, Programmed by Eli Bachmutsky, Copyleft Feb.1999" */
	/* rules: 65, states: 8, format: CNESWC' */
	static int[] chou2TransitionTable = {
		0000000, 0000440, 0000547, 0000100, 0000110,
		0000330, 0004040, 0004445, 0004100, 0001040,
		0001010, 0001740, 0003000, 0003010, 0003030,
		0007040, 0007030, 0007104, 0007110, 0040000,
		0040070, 0047100, 0050007, 0017000, 0070000,
		0070077, 0071010, 0400101, 0400313, 0401033,
		0407103, 0411033, 0431033, 0500033, 0503330,
		0100045, 0100011, 0100414, 0101044, 0101301,
		0103011, 0107131, 0140414, 0141044, 0111044,
		0133011, 0177711, 0304011, 0305011, 0305141,
		0307141, 0344011, 0345011, 0354010, 0314001,
		0377711, 0700000, 0700337, 0707100, 0707141,
		0707110, 0717000, 0730007, 0770070, 0770711,
	};

	/* From Hiroki Sayama */
	static int[] evoTransitionTable = {
		0000012, 0000043, 0000122, 0000152, 0000212,
		0000242, 0000422, 0000452, 0000752, 0001022,
		0002141, 0002171, 0002322, 0011221, 0012121,
		0012321, 0012421, 0012451, 0012526, 0012626,
		0012721, 0012751, 0013421, 0013721, 0014221,
		0014251, 0014321, 0014351, 0014421, 0014621,
		0017221, 0017251, 0017561, 0017621, 0017721,
		0100011, 0100121, 0100211, 0100244, 0100277,
		0101211, 0101244, 0101277, 0102021, 0102111,
		0102121, 0102131, 0102211, 0102244, 0102277,
		0102324, 0102414, 0102424, 0102434, 0102511,
		0102527, 0102543, 0102577, 0102717, 0102727,
		0102735, 0105121, 0105424, 0105727, 0106211,
		0106244, 0106277, 0111121, 0111221, 0111244,
		0111251, 0111277, 0111621, 0112121, 0112131,
		0112151, 0112221, 0112244, 0112277, 0112321,
		0112424, 0112434, 0112527, 0112543, 0112577,
		0112626, 0112727, 0112735, 0113221, 0113321,
		0115424, 0115727, 0116244, 0116277, 0122244,
		0122277, 0122434, 0122737, 0123244, 0123277,
		0124266, 0124333, 0126276, 0200012, 0200022,
		0200042, 0200052, 0200060, 0200071, 0200122,
		0200152, 0200212, 0200222, 0200232, 0200242,
		0200260, 0200272, 0200324, 0200423, 0200452,
		0200545, 0200575, 0200620, 0200722, 0200752,
		0201022, 0201122, 0201222, 0201422, 0201722,
		0202022, 0202032, 0202052, 0202065, 0202073,
		0202122, 0202152, 0202212, 0202222, 0202232,
		0202323, 0202422, 0202452, 0202525, 0202620,
		0202650, 0202722, 0202752, 0203122, 0203222,
		0203422, 0203452, 0203722, 0204122, 0204222,
		0204422, 0205122, 0205425, 0205725, 0206125,
		0206212, 0206425, 0206725, 0207122, 0207222,
		0207722, 0211222, 0212222, 0212232, 0212242,
		0212272, 0212323, 0213222, 0214222, 0216222,
		0217222, 0222242, 0222272, 0222342, 0222372,
		0222432, 0222442, 0222732, 0222772, 0223243,
		0223273, 0300013, 0300022, 0300032, 0300043,
		0300074, 0300123, 0300322, 0300421, 0301021,
		0301250, 0302123, 0302423, 0302521, 0302723,
		0303321, 0312123, 0312423, 0312521, 0312723,
		0324243, 0324251, 0324273, 0325271, 0327273,
		0400001, 0400021, 0401020, 0401120, 0401220,
		0401250, 0401620, 0402120, 0402150, 0402221,
		0402321, 0402626, 0403120, 0403221, 0500025,
		0500125, 0500215, 0500232, 0500245, 0500275,
		0500425, 0500725, 0502022, 0502052, 0502125,
		0502152, 0502425, 0502725, 0503120, 0602022,
		0602122, 0602220, 0602422, 0602722, 0612220,
		0622240, 0622270, 0701020, 0701120, 0701220,
		0701250, 0701620, 0702120, 0702150, 0702221,
		0702320, 0702626, 0703120
	};

	/* From David Bagley */
	static int[] trihexTransitionTable = { /* Octal CNESswn->I */
	  /* CNESswnI   CNESswnI   CNESswnI   CNESswnI   CNESswnI */
	/* TRIA -- experimental */
	    000000000, 000000020, 000000220, 000002220, 000022220,
	    011122121, 011121221, 011122221, 011221221,
	    011222221, 011112121, 011112221,
	    020021122, 020002122, 020211222, 021111222,
	    020221122, 020027122, 020020722, 020021022,
	    001127221,
	    011122727, 011227227, 010122121, 010222211,
	    021117222, 020112272,
	    070221220,
	    001227221,
	    010221121, 011721221, 011222277,
	    020111222, 020221172,
	    070211220,
	    001217221,
	    010212277, 010221221,
	    020122112,
	    070122220,
	    001722221,
	    010221271,
	    020002022, 021122172,
	    070121220,
	    011122277, 011172121,
	    010212177, 011212277,
	    070112220,
	    001772221,
	    021221772,
	    070121270, 070721220,
	    000112721, 000272211,
	    010022211, 012222277,
	    020072272, 020227122, 020217222,
	    010211121,
	    020002727,
	    070222220,
	    001727721,
	    020021072, 020070722,
	    070002072, 070007022,
	    001772721,
	    070002022,
	    000000070, 000000770, 000072220, 000000270,
	    020110222, 020220272, 020220722,
	    070007071, 070002072, 070007022,
	    000000012, 000000122, 000000212, 001277721,
	    020122072, 020202212,
	    010002121,
	    020001122, 020002112,
	    020021722,
	    020122022, 020027022, 020070122, 020020122,
	    010227027,
	    020101222,
	    010227227, 010227277,
	    021722172,
	    001727221,
	    010222277,
	    020702272,
	    070122020,
	    000172721,
	    010022277, 010202177, 010227127,

	    001214221,
	    010202244,
	    020024122, 020020422,
	    040122220,
	    001422221,
	    010221241, 010224224,
	    021122142,
	    040121220,
	    001124221,
	    010224274,
	    020112242, 021422172,
	    040221220,
	    001224221, 001427221,
	    010222244,
	    020227042,
	    040122020,
	    000142721,
	    010022244, 010202144, 010224124,
	    040112220,
	    001442221,
	    021221442,
	    040121240, 040421220,
	    000242211, 000112421,
	    020042242, 020214222, 020021422, 020220242, 020024022,
	    011224224,
	    020224122,
	    020220422,
	    012222244,
	    020002424,
	    040222220,
	    001244421, 000000420, 000000440, 000000240, 000000040,
	    020040121, 020021042,
	    040004022, 040004042, 040002042,
	    010021121,
	    020011122, 020002112,
	    001424421,
	    020040422,
	    001442421,
	    040002022,
	    001724221,
	    010227247,
	    020224072, 021417222,
	    000172421,
	    010021721,
	    020017022,
	    020120212,
	    020271727,
	    070207072, 070701220,
	    000001222,
	    020110122,
	    001277221,
	    001777721,
	    020021222, 020202272, 020120222, 020221722,
	    020027227,
	    070070222,
	    000007220,
	    020101272, 020272172, 020721422, 020721722,
	    020011222, 020202242,
	};

	 /* Hex */
	static int[] hexTransitionTable = { /* Octal CNESswn->I */
	  /* CNESswnI   CNESswnI   CNESswnI   CNESswnI   CNESswnI */
	    000000000, 000000020, 000000220, 000002220,
	    011212121, 011212221, 011221221, 011222221,
	    020002122, 020021122, 020211122,

	    010221221, 010222121,
	    020002022, 020021022, 020020122, 020112022,

	    010202121,
	    020102022, 020202112,

	    000000012, 000000122, 000000212,
	    010002121,
	    020001122, 020002112, 020011122,


	    001227221, 001272221, 001272721,
	    012212277, 011222727, 011212727,
	    020021722, 020027122, 020020722, 020027022,
	    020211722, 020202172, 020120272,
	    020271122, 020202172, 020207122, 020217122,
	    020120272, 020210722, 020270722,
	    070212220, 070221220, 070212120,


	    012222277,
	    020002727,
	    070222220,

	    001277721, 000000070, 000000270, 000000720, 000000770,
	    020070122, 020021072,
	    070002072, 070007022, 070007071,

	    020070722,
	    070002022,

	    010227227, 010222727, 010202727,
	    020172022, 020202712,

	    001224221, 001242221, 001242421,
	    012212244, 011222424, 011212424,
	    020021422, 020024122, 020020422, 020024022,
	    020211422, 020202142, 020120242,
	    020241122, 020202142, 020204122, 020214122,
	    020120242, 020210422, 020240422,
	    040212220, 040221220, 040212120,


	    012222244,
	    020002424,
	    040222220,

	    001244421, 000000040, 000000240, 000000420, 000000440,
	    020040122, 020021042,
	    040002042,
	    040004021, 040004042,

	    020040422,
	    040002022,

	    010224224, 010222424, 010202424,
	    020142022, 020202412,
	    020011722, 020112072, 020172072, 020142072,



	    000210225, 000022015, 000022522,
	    011225521,
	    020120525, 020020152, 020005122, 020214255, 020021152,
	    020255242,
	    050215222, 050225121,

	    000225220, 001254222,
	    010221250, 011221251, 011225221,
	    020025122, 020152152, 020211252, 020214522, 020511125,
	    050212241, 05221120,
	    040521225,

	    000000250, 000000520, 000150220, 000220520, 000222210,
	    001224251,
	    010022152, 010251221, 010522121, 011212151, 011221251,
	    011215221,
	    020000220, 020002152, 020020220, 020022152,
	    020021422, 020022152, 020022522, 020025425, 020050422,
	    020051022, 020051122, 020211122, 020211222, 020215222,
	    020245122,
	    050021125, 050021025, 050011125, 051242221,
	    041225220,

	    000220250, 000220520, 001227521, 001275221,
	    011257227, 011522727,
	    020002052, 020002752, 020021052, 020057125,
	    050020722, 050027125,
	    070215220,

	    070212255,
	    071225220,
	    020275122,
	    051272521,
	    020055725,
	    020021552,
	    012252277,
	    050002521,
	    020005725,

	    050011022,
	    000000155,
	    020050722,
	    001227250,
	    010512727,
	    010002151,
	    020027112,
	    001227251,
	    012227257,
	    050002125,
	    020517122,
	    050002025,
	    020050102,
	    050002725,
	    020570722,
	    001252721,
	    020007051,
	    020102052,
	    020271072,
	    050001122,
	    010002151,
	    011227257,
	    020051722,
	    020057022,
	    020050122,


	    020051422,
	    011224254,
	    012224254,

	    020054022,
	    050002425,
	    040252220,
	    020002454,


	    000000540,
	    001254425,
	    050004024,
	    040004051,

	    000000142,
	    040001522,
	    010002547,
	    020045122,
	    051221240,
	    020002512,
	    020021522,


	    020020022,
	    021125522,
	    020521122,
	    020025022,
	    020025522,
	    020020522,

	    020202222,
	    020212222,
	    021212222,
	    021222722,
	    021222422,
	    020002222,
	    020021222,
	    020022122,
	    020212122,
	    020027222,
	    020024222,
	    020212722,
	    020212422,
	    020202122,
	    001222221,
	    020002522,

	    020017125,
	    010022722,
	    020212052,

	    020205052,
	    070221250,

	    000000050, 000005220, 000002270, 070252220,
	    000000450, 000007220,
	    000220220, 000202220, 000022020, 000020220,

	    000222040,
	    000220440,
	    000022040,
	    000040220,

	    000252220,
	    050221120, 010221520,
	    002222220,

	    000070220, 000220720,
	    000020520, 000070250, 000222070, 000027020,
	    000022070, 000202270, 000024020, 000220420,
	    000220270, 000220240, 000072020, 000042020,
	    000002020, 000002070, 000020270, 000020250,
	    000270270, 000007020, 000040270,

	    /* Collision starts (gen 540), not sure to have rules to save it
	       or depend on calloc to initialize remaining rules to 0 so that
	       the mutant will be born  (mutant develops 2 arms at 1176)
	     */
	    000050220,
	};

	static final String[] defaultRuleStrings = {
		//"",
		"Langton's Loops",
		//"Byl's Loops",
		//"Chou Reggia 1 Loops",
		//"Chou Reggia 2 Loops",
		//"Evolving Loops",
		//"Disolving Loops",
		//"Disolving Sheath Loops",
		//"Sheath Loops",
		//"",
		"Hexagonal Loops"
	};

	static final Integer[] defaultRuleIndex = {
		//null,
		//null,
		0,
		//null,
		0};

	static final LoopRules[][] loopRules = {
		//{ /* 3 Triangular, von Neuman */ },
		{ /* 4 Square, von Neuman */
			new LoopRules("Langton's Loops", "Chris Langton"),
			new LoopRules("Byl Loops", "J. Byl."),
			new LoopRules("Chou Reggia 1 Loops", "Chou, Reggia"),
			new LoopRules("Chou Reggia 2 Loops", "Chou, Reggia"),
			new LoopRules("Evolving Loops", "Hiroki Sayama"),
			new LoopRules("Disolving Loops", "Hiroki Sayama"),
			new LoopRules("Disolving Sheath Loops", "Hiroki Sayama"),
			new LoopRules("Sheath Loops", "Hiroki Sayama"),
		},
		//{ /* 5 Pentagonal, Cairo Tiling */ },
		{ /* 6 Hexagonal */
			new LoopRules("Hexagonal Loops", "David Bagley"),
		}
	};

	// Array of Generic Hash Tables
	ArrayList<HashMap<String, Integer>> arrayHashMap = new ArrayList<>();
//http://forums.sun.com/thread.jspa?threadID=530823
	final void init() {
		if (!arrayHashMap.isEmpty())
			arrayHashMap.clear();
		for (int i = 0; i < MAX_NEIGHBOR_INDEXES; i++) {
//System.out.println("Loop.init: " + i);
			arrayHashMap.add(new HashMap<String, Integer>());
//System.out.println("Loop.init: " + neighborIndex + " " + loopRules[neighborIndex].length);
			for (int j = 0; j < loopRules[i].length; j++) {
//System.out.println("Loop.init: " + indexToNeighbor[neighborIndex] + loopRules[neighborIndex][i].ruleName + " " + i);
				arrayHashMap.get(i).put(loopRules[i][j].ruleName, j);
			}
		}
	}

	Integer getRuleIndex(int neighborhoodIndex, String s) {
		if (loopRules[neighborhoodIndex].length == 0)
			return null;
//System.out.println("Loop getRuleIndex: value " + arrayHashMap.get(neighborIndex).values());
//System.out.println("Loop.getRuleIndex: key " + arrayHashMap.get(neighborIndex).keySet());
//System.out.println("Loop.getRuleIndex: n " + indexToNeighbor[neighborIndex] + " " + s + " " + arrayHashMap.get(neighborIndex).get(s));
		return arrayHashMap.get(neighborhoodIndex).get(s);
	}

	static String[] getStrings(int neighborIndex) {
		String[] strings = new String[loopRules[neighborIndex].length];
		int i;

		for (i = 0; i < loopRules[neighborIndex].length; i++) {
			strings[i] = loopRules[neighborIndex][i].ruleName;
			if (!"".equals(loopRules[neighborIndex][i].discoverer))
				strings[i] += ", " + loopRules[neighborIndex][i].discoverer;
		}
		return strings;
		//return (String[]) (arrayHashMap.get(neighborIndex).values()).toArray();
		//Collection<String> strings = arrayHashMap.get(neighborIndex).values();
		//return (String[]) strings.toArray();
		//String[] strings = new String[arrayHashMap.size()];
		//for (int i = 0; i < loopRules[ni].length; i++) {
	//		strings[i] = arrayHashMap.get(neighborIndex).getKey(i);
		//}
		//return strings;
	}

	static final int[] neighborIndexToPolygon = {
		3, 4, 5, 6,
	};

	static int neighborToPolygon(int neighbors) {
		switch (neighbors) {
		//case 3: /* triangle */
		//	return 3;
		case 4: /* square */
			return 4;
		//case 5: /* pentagon, Cairo Tiling */
		//	return 5;
		case 6: /* hexagon */
			return 6;
		default:
			System.out.println("no polygon known for " +
				neighbors + " neighbors");
			return 4;
		}
	}

	static final String[] neighborIndexStrings = {
		//"3, Triangular, von Neuman",
		"4, Square, von Neuman",
		//"5, Pentagon, Cairo Tiling",
		"6, Hexagonal"
	};

	static final int[] indexToNeighbor = {
		//3,
		4,
		//5,
		6,
	};

	static int neighborToIndex(int neighbors) {
		switch (neighbors) {
		//case 3: /* von Neuman Neighbor Count */
		//	return 0;
		case 4: /* von Neuman Neighbor Count */
			return 0;
		//case 5: /* Cario Tiling */
		//	return 2;
		case 6:
			return 1;
		default:
			System.out.println("no neighborhood like " +
				neighbors + " known");
			return LoopInterface.DEFAULT_NEIGHBOR_INDEX;
		}
	}

	static String parse(String rule) {
		return indexToNeighbor[parseNeighborFromRule(rule)] +
			parseNeighborRule(rule);
	}

	static int parseNeighborFromRule(String rule) {
		int neighbor = 0;
		int i = 0;
		boolean negative = false;

		for (; i < rule.length(); i++) {
			char letter = rule.charAt(i);

			if (letter >= '0' && letter <= '9')
				neighbor = neighbor * BASE + (letter - '0');
			else if (letter == '-')
				negative = !negative;
			else
				break;
		}
		if (negative)
			return neighborToIndex(-neighbor);
		return neighborToIndex(neighbor);
	}

	static String parseNeighborRule(String rule) {
		int neighborIndex = parseNeighborFromRule(rule);

		return defaultRuleStrings[neighborIndex];
	}

	  /* square: 8*8*8*8      = 2^12 = 2^3^4 = 4096 */
	  /* extsquare: 9*9*9*9   = 3^8 = 3^2^4 = 6561 */
	  /* inefsquare: 9*9*9*9*9   = 3^10 = 3^2^5 = 59049 */
	  /* hexagon: 8*8*8*8*8*8 = 2^18 = 2^3^6 = 262144 = too big? */
	static final int[] M6 = {1, 6, 36, 216, 1296, 7776, 46656, 279936, 1679616};
	static final int[] M8 = {1, 8, 64, 512, 4096, 32768, 262144, 2097152, 16777216};
	static final int[] M9 = {1, 9, 81, 729, 6561, 59049, 531441, 4782969, 43046721,
		387420489};
	int[] table = null;
	static int factor = 8;
	static int factor2 = factor * factor;
	static int factor3 = factor2 * factor;
	static int factor4 = factor3 * factor;


	private int inefTableOut(int c, int w, int s, int e, int n) {
		return (table[n*factor4+e*factor3+s*factor2+w*factor+c]);
	}

	private void inefTableIn(int c, int w, int s, int e, int n, int i) {
		table[n*factor4+e*factor3+s*factor2+w*factor+c] = i;
	}

	private void setGeneralTransition(int tt) {
		int i, c, k, cond;
		int neighbors = getNeighbors();
		int[] n = new int[neighbors];
		/* zeroed out rules used here...
			also could use static rules
			also could use random rules
		   Instead of setting "unused" state rules to zero it randomizes them.
		   These rules take over when something unexpected happens... like when a
		   cell hits a wall (the end of the screen).
		 */

		i = tt % 8;
		cond = tt / 8;
		for (k = 0; k < neighbors; k++) {
			n[k] = cond % 8;
			cond /= 8;
		}
		c = cond % 8;
		if (neighbors == 6) { /* (factor == 8) */
			for (k = 0; k < neighbors; k++) {
				table[n[k]<<15|
					n[(k + 1) % neighbors]<<12|
					n[(k + 2) % neighbors]<<9|
					n[(k + 3) % neighbors]<<6|
					n[(k + 4) % neighbors]<<3|
					n[(k + 5) % neighbors]]|=(i<<(c*3));
			}
		} else if (neighbors == 4) {
			if (factor == 8) {
				for (k = 0; k < neighbors; k++) {
					table[n[k]<<9|
						n[(k + 1) % neighbors]<<6|
						n[(k + 2) % neighbors]<<3|
						n[(k + 3) % neighbors]]|=(i<<(c*3));
				}
			} else {
				factor2 = factor * factor;
				factor3 = factor2 * factor;
				factor4 = factor3 * factor;
				for (k = 0; k < neighbors; k++) {
					table[n[k]*factor4+
						n[(k + 1) % neighbors]*factor3+
						n[(k + 2) % neighbors]*factor2+
						n[(k + 3) % neighbors]*factor+c] = i;
				}
			}
		}
	}

	byte getTransition(byte c, int[] n, boolean clockwise) {
		int neighbors = getNeighbors();

		if (n == null)
			return 0;
		if (neighbors == 6) // factor == 8
			return (byte) ((clockwise) ?
				(table[n[0]<<15|n[1]<<12|n[2]<<9|n[3]<<6|n[4]<<3|n[5]]>>(c*3))&7 :
				(table[n[5]<<15|n[4]<<12|n[3]<<9|n[2]<<6|n[1]<<3|n[0]]>>(c*3))&7);
		else if (neighbors == 4) {
			if (factor == 8) {
				return (byte) ((clockwise) ?
					(table[n[0]<<9|n[1]<<6|n[2]<<3|n[3]]>>(c*3))&7 :
					(table[n[3]<<9|n[2]<<6|n[1]<<3|n[0]]>>(c*3))&7);
			} else if (dissolve || evolve || sheath) {
				return (byte) ((clockwise) ?
					(table[n[0]*factor4+n[1]*factor3+n[2]*factor2+n[3]*factor+c]) :
					(table[n[3]*factor4+n[2]*factor3+n[1]*factor2+n[0]*factor+c]));
			}
		}
		return 0;
	}

	static int sheathCount(int a) {
		return (((a==1)||(a==2)||(a==4)||(a==6)||(a==7)) ? 1 : 0);
	}

	static boolean inSheath(int b, int c, int d, int e) {
		return ((sheathCount(b) + sheathCount(c) +
			sheathCount(d) + sheathCount(e)) >= 2);
	}

	void setInitRule() {
		int a, b, c, d, e;
		final int extStates = 9;

		for (a = 0; a < extStates; a++)
		for (b = 0; b < extStates; b++)
		for (c = 0; c < extStates; c++)
		for (d = 0; d < extStates; d++)
		for (e = 0; e < extStates; e++)
		inefTableIn(a, b, c, d, e, extStates);
	}

	void setUndefinedRule() {
		int a, b, c, d, e;
		final int states = 8;
		final int extStates = 9;

		if (sheath) {
			for (a = 0; a < states; a++)
			for (b = 0; b < states; b++)
			for (c = 0; c < states; c++)
			for (d = 0; d < states; d++)
			for (e = 0; e < states; e++)
			if (inefTableOut(a, b, c, d, e) == extStates)
				switch(a) {
				default:
				case 0:
					if (((b == 1) || (c == 1) ||
							(d == 1) || (e == 1)) &&
							(inSheath(b, c, d, e))) {
						inefTableIn(a, b, c, d, e, 1);
					} else {
						inefTableIn(a, b, c, d, e, 0);
					}
					break;
				case 1:
					if (((b == 7) || (c == 7) ||
							(d == 7) || (e == 7)) &&
							(inSheath(b, c, d, e))) {
						inefTableIn(a, b, c, d, e, 7);
					} else if (((b == 6) || (c == 6) ||
							(d == 6) || (e == 6)) &&
							(inSheath(b, c, d, e))) {
						inefTableIn(a, b, c, d, e, 6);
					} else if (((b == 4) || (c == 4) ||
							(d == 4) || (e == 4)) &&
							(inSheath(b, c, d, e))) {
						inefTableIn(a, b, c, d, e, 4);
					}
					break;
				case 2:
					if (((b == 3) || (c == 3) ||
							(d == 3) || (e == 3))) {
						inefTableIn(a, b, c, d, e, 1);
					} else if (((b == 2) || (c == 2) ||
							(d == 2) || (e == 2))) {
						inefTableIn(a, b, c, d, e, 2);
					}
					break;
				case 4:
				case 6:
				case 7:
					if (((b == 0) || (c == 0) ||
							(d == 0) || (e == 0)) &&
							(inSheath(b, c, d, e))) {
						inefTableIn(a, b, c, d, e, 0);
					}
					break;
				}
		}
		if (dissolve) {
			for (a = 0; a < extStates; a++)
			for (b = 0; b < extStates; b++)
			for (c = 0; c < extStates; c++)
			for (d = 0; d < extStates; d++)
			for (e = 0; e < extStates; e++)
			if (inefTableOut(a, b, c, d, e) == extStates) {
				if (a == 8) {
					inefTableIn(a, b, c, d, e, 0);
				} else if ((b == 8) || (c == 8) ||
						(d == 8) || (e == 8))
					switch(a) {
					default:
					case 0:
					case 1:
						if (((b > 1) && (b < 8)) ||
								((c > 1) && (c < 8)) ||
								((d > 1) && (d < 8)) ||
								((e > 1) && (e < 8))) {
							inefTableIn(a, b, c, d, e, 8);
						} else {
							inefTableIn(a, b, c, d, e, a);
						}
						break;
					case 2:
					case 3:
					case 5:
						inefTableIn(a, b, c, d, e, 0);
						break;
					case 4:
					case 6:
					case 7:
						inefTableIn(a, b, c, d, e, 1);
						break;
					}
			}
			inefTableIn(1, 1, 1, 5, 2, 8);
			inefTableIn(1, 1, 5, 2, 1, 8);
			inefTableIn(1, 5, 2, 1, 1, 8);
			inefTableIn(1, 2, 1, 1, 5, 8);
		}
		/* Clearance of undefined rules */
		for (a = 0; a < extStates; a++)
		for (b = 0; b < extStates; b++)
		for (c = 0; c < extStates; c++)
		for (d = 0; d < extStates; d++)
		for (e = 0; e < extStates; e++)
		if (inefTableOut(a, b, c, d, e) == extStates) {
			if (dissolve) {
				inefTableIn(a, b, c, d, e, 8);
			} else {
				inefTableIn(a, b, c, d, e, a);
			}
		}
	}

	void setUndefinedEvoRule() {
		int a, b, c, d, e;
		final int extStates = 9;

		for (a = 0; a < extStates; a++)
		for (b = 0; b < extStates; b++)
		for (c = 0; c < extStates; c++)
		for (d = 0; d < extStates; d++)
		for (e = 0; e < extStates; e++)
		if (inefTableOut(a, b, c, d, e) == extStates) {
			if (a == 8) {
				inefTableIn(a, b, c, d, e, 0);
			} else if ((b == 8) || (c == 8) || (d == 8) || (e == 8))
				switch(a) {
				default:
				case 0:
				case 1:
					if (((b > 1) && (b < 8)) ||
							((c > 1) && (c < 8)) ||
							((d > 1) && (d < 8)) ||
							((e > 1) && (e < 8))) {
						inefTableIn(a, b, c, d, e, 8);
					} else {
						inefTableIn(a, b, c, d, e, a);
					}
					break;
				case 2:
				case 3:
				case 5:
					inefTableIn(a, b, c, d, e, 0);
					break;
				case 4:
				case 6:
				case 7:
					inefTableIn(a, b, c, d, e, 1);
					break;
				}
		}
		/* Clearance of undefined rules */
		for (a = 0; a < extStates; a++)
		for (b = 0; b < extStates; b++)
		for (c = 0; c < extStates; c++)
		for (d = 0; d < extStates; d++)
		for (e = 0; e < extStates; e++)
		if (inefTableOut(a, b, c, d, e) == extStates) {
			if (a == 0) {
				inefTableIn(a, b, c, d, e, 0);
			} else {
				inefTableIn(a, b, c, d, e, 8);
			}
		}
	}

	void initTable(int neighbors, int ruleIndex) {
		int j, length;

		//System.out.println("initTable:" + neighbors + ", " + ruleIndex);
		if (neighbors == 4) {
			dissolve = false;
			sheath = false;
			switch (ruleIndex) {
			case 0:
				table = new int[4096];
				length = transitionTable.length;
				// static rules
				for (j = 0; j < length; j++)
					setGeneralTransition(transitionTable[j]);
				break;
			case 1:
				table = new int[4096];
				length = bylTransitionTable.length;
				for (j = 0; j < length; j++)
					setGeneralTransition(bylTransitionTable[j]);
				break;
			case 2:
				table = new int[4096];
				length = chou1TransitionTable.length;
				for (j = 0; j < length; j++)
					setGeneralTransition(chou1TransitionTable[j]);
				break;
			case 3:
				table = new int[4096];
				length = chou2TransitionTable.length;
				for (j = 0; j < length; j++)
					setGeneralTransition(chou2TransitionTable[j]);
				break;
			default:
				if (ruleIndex == 4)
					evolve = true;
				else if (ruleIndex == 5 || ruleIndex == 6)
					dissolve = true;
				if (ruleIndex == 6 || ruleIndex == 7)
					sheath = true;
				table = new int[59049];
				setInitRule();
				if (!(dissolve || evolve || sheath))
					System.out.println("initTable: error");
				factor = 9;
				length = (evolve) ? evoTransitionTable.length :
					transitionTable.length;
				for (j = 0; j < length; j++)
				if (evolve)
					setGeneralTransition(evoTransitionTable[j]);
				else
					setGeneralTransition(transitionTable[j]);
				if (evolve)
					setUndefinedEvoRule();
				else
					setUndefinedRule();
				break;
			}
		} else if (neighbors == 6) {
			length = hexTransitionTable.length;
			table = new int[262144];
			dissolve = false;
			sheath = false;
			for (j = 0; j < length; j++)
				setGeneralTransition(hexTransitionTable[j]);
		}
	}

	// for debugging
	static String toString(int neighbors, boolean verbose) {
		StringBuilder sb = new StringBuilder("Rule");

		if (verbose)
			System.out.print("rule (Survival/Birth " +
				neighbors + " neighborhood): ");
		//int neighborIndex = neighborToIndex(neighbors);
		if (verbose) {
			System.out.println(sb);
		}
		return sb.toString();
	}
}
