package com.sillycycle.bagleyd.util;

/*
 * @(#)CellSq.java
 *
 * Copyright 2008 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

/**
 * The <code>CellSq</code> class handles
 * the construction of the square cells.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class CellSq extends Component {
	private static final long serialVersionUID = 42L;

	Rectangle r;

	CellSq(int nx, int ny, int screenWidth, int screenHeight, boolean scroll) {
		int offset;
		int startX = 0, startY = 0;
		int xs, ys;

		if (scroll) {
			offset = 16;
			xs = (screenWidth - offset) / nx;
			ys = (screenHeight - offset) / ny;
		} else {
			offset = 6;
			xs = (screenWidth - offset) / nx;
			ys = (screenHeight - offset) / ny;
			startX = (screenWidth - offset - xs * nx) / 2;
			startY = (screenHeight - offset - ys * ny) / 2;
		}
		if (xs < 1)
			xs = 1;
		if (ys < 1)
			ys = 1;
		r = new Rectangle(startX, startY, xs, ys);
	}

	Rectangle boundRect(int nx, int ny) {
		return new Rectangle(r.x - 2, r.y - 2,
			r.x + r.width * nx + 3,
			r.y + r.height * ny + 3);
	}

	Point screenToCell(int x, int y) {
		return new Point(ModArith.div(x - r.x, r.width),
			ModArith.div(y - r.y, r.height));
	}

	void drawCell(Graphics g, boolean round, Color cid, Color bid,
			int i, int j) {
		if (g == null)
			return;
		if (cid != null) {
			g.setColor(cid);
			if (round) {
				g.fillOval(r.x + r.width * i + 1,
					r.y + r.height * j + 1,
					r.width - 1, r.height - 1);
			} else {
				g.fillRect(r.x + r.width * i + 1,
					r.y + r.height * j + 1,
					r.width - 1, r.height - 1);
			}
		}
		if ((bid != null || cid == null) &&
				r.width > 1 && r.height > 1 ) {
			if (bid != null)
				g.setColor(bid);
			g.drawRect(r.x + r.width * i,
				r.y + r.height * j,
				r.width, r.height);
		}
	}

	void drawCaptureRect(Graphics g, Color color,
			int i, int j, int nx, int ny) {
		if (g == null)
			return;
		g.setColor(color);
		/* Just drawing a rectangle is fine but
		 * for other grids this is hard to do.
		 * Change for all grids for consistency.
		 * LCD unfortunately. */
		/*g.drawRect(r.x + nx * i, y + ny * j,
			r.width * nx, r.height * ny);*/
		for (int vert = j; vert < ny + j; vert++) {
			for (int horiz = i; horiz < nx + i; horiz++) {
				drawCell(g, false, null, null, horiz, vert);
			}
		}
	}

	void drawEyes(Graphics g, Color color, short direction,
			int i, int j) {
		// so far this does not scale, as its always a pixel
		/* if small return */
		if (!(r.width > 3 && r.height > 3))
			return;
		int offsetX = r.x + 1;
		int offsetY = r.y + 1;

		g.setColor(color);
		switch (direction) {
		case 0:
			g.fillRect(
				offsetX + r.width * (i + 1) - 4,
				offsetY + r.height * j + r.height / 2 - 3,
				2, 2);
			g.fillRect(
				offsetX + r.width * (i + 1) - 4,
				offsetY + r.height * j + r.height / 2 + 1,
				2, 2);
			break;
		case 45:
			g.fillRect(
				offsetX + r.width * (i + 1) - 5,
				offsetY + r.height * j,
				2, 2);
			g.fillRect(
				offsetX + r.width * (i + 1) - 3,
				offsetY + r.height * j + 2,
				2, 2);
			break;
		case 90:
			g.fillRect(
				offsetX + r.width * i + r.width / 2 - 3,
				offsetY + r.height * j + 1,
				2, 2);
			g.fillRect(
				offsetX + r.width * i + r.width / 2 + 1,
				offsetY + r.height * j + 1,
				2, 2);
			break;
		case 135:
			g.fillRect(
				offsetX + r.width * i + 3,
				offsetY + r.height * j,
				2, 2);
			g.fillRect(
				offsetX + r.width * i + 1,
				offsetY + r.height * j + 2,
				2, 2);
			break;
		case 180:
			g.fillRect(
				offsetX + r.width * i + 2,
				offsetY + r.height * j + r.height / 2 - 3,
				2, 2);
			g.fillRect(
				offsetX + r.width * i + 2,
				offsetY + r.height * j + r.height / 2 + 1,
				2, 2);
			break;
		case 225:
			g.fillRect(
				offsetX + r.width * i + 2,
				offsetY + r.height * (j + 1) - 3,
				2, 2);
			g.fillRect(
				offsetX + r.width * i,
				offsetY + r.height * (j + 1) - 5,
				2, 2);
			break;
		case 270:
			g.fillRect(
				offsetX + r.width * i + r.width / 2 - 3,
				offsetY + r.height * (j + 1) - 4,
				2, 2);
			g.fillRect(
				offsetX + r.width * i + r.width / 2 + 1,
				offsetY + r.height * (j + 1) - 4,
				2, 2);
			break;
		case 315:
			g.fillRect(
				offsetX + r.width * (i + 1) - 5,
				offsetY + r.height * (j + 1) - 3,
				2, 2);
			g.fillRect(
				offsetX + r.width * (i + 1) - 3,
				offsetY + r.height * (j + 1) - 5,
				2, 2);
			break;
		default:
			System.out.println("wrong eyes direction " +
				direction + " for ant eyes");
		}
	}

	void drawTruchet(Graphics g, Color color, int truchetState,
			int i, int j) {
		g.setColor(color);
		int offsetX = r.x + 1;
		int offsetY = r.y + 1;
		if (truchetState == 0) {
			g.drawArc(
				offsetX + r.width * i - (r.width - 1) / 2,
				offsetY + r.height * j - (r.height - 1) / 2,
				r.width - 2, r.height - 2,
				-1, -88);
			g.drawArc(
				offsetX + r.width * i + (r.width - 1) / 2,
				offsetY + r.height * j + (r.height - 1) / 2,
				r.width - 2, r.height - 2,
				91, 88);
		} else {
			g.drawArc(
				offsetX + r.width * i - (r.width - 1) / 2,
				offsetY + r.height * j + (r.height - 1) / 2,
				r.width - 2, r.height - 2,
				1, 88);
			g.drawArc(
				offsetX + r.width * i + (r.width - 1) / 2,
				offsetY + r.height * j - (r.height - 1) / 2 + 1,
				r.width - 2, r.height - 2,
				-91, -88);
		}
		/*g.setColor(bid);
		g.drawRect(r.x + r.width * i,
			r.height * j,
			r.width, r.height);*/ // fix drawing errors
	}
}
